"""
The module with CLI code that handles the `hopla request` command.
"""
import logging
from typing import List, Tuple, Optional

import click
import requests
from hopla.hoplalib.http import RequestHeaders
from hopla.hoplalib.outputformatter import JsonFormatter

supported_http_request_methods = click.Choice(["GET", "POST"])

log = logging.getLogger()

BodyParam = Tuple[str, str]


@click.command()
@click.option("--method", "-X", default="GET", show_default=True,
              type=supported_http_request_methods,
              help="The HTTP method.")
@click.option("--domain", default="https://habitica.com", show_default=True,
              metavar="URL",
              help="The domain name where the Habitica API is hosted.")
@click.option("--body-param", "body_param_list", multiple=True, type=click.Tuple([str, str]),
              metavar="KEY VALUE",
              help="A key-value pair for in the JSON body. E.g. --body-param 'dayStart' '0' ."
                   "The --body-param option can be used multiple times.")
@click.argument("path")
def request(method: str,
            domain: str,
            body_param_list: Optional[List[BodyParam]],
            path: str):
    """Perform a HTTP request on the Habitica API.

    PATH This is the path of the endpoint that you want to perform a HTTP
    request on. For example, /api/v3/groups.

    \b
    Examples
    ----
    # GET: get party information
    $ hopla request /api/v3/groups/party

    \b
    # GET: a user's tasks
    $ hopla request /api/v3/tasks/user

    \b
    # GET: a user's habits
    $ hopla request /api/v3/tasks/user?type=habits

    \b
    # POST: open a mystery item
    $ hopla request --method=POST /api/v3/user/open-mystery-item

    \b
    # POST: send a quest invite for Recidivate pt. 3 to your party
    $ hopla request --method POST /api/v3/groups/party/quests/invite/moonstone3

    \b
    # POST: buy a special item. In this case, a spooky sparkles item.
    $ hopla request -XPOST /api/v3/user/buy-special-spell/spookySparkles

    \b
    # POST: cast the Searing Brightness (Healer) and Tool of Trade (Rogue) spell.
    $ hopla request -XPOST /api/v3/user/class/cast/brightness
    $ hopla request -XPOST /api/v3/user/class/cast/toolsOfTrade

    \b
    # POST: set your custom day start to 1 AM
    $ hopla request -XPOST --body-param dayStart 1 /api/v3/user/custom-day-start

    \f
    :return:
    """
    log.debug(f"hopla request {method} DOMAIN+PATH={domain + path}")
    log.debug(f"   body_param_list={body_param_list}")

    request_endpoint: str = domain + path
    headers: dict = RequestHeaders().get_default_request_headers()

    # no support for query parameters yet
    if body_param_list:
        http_request = requests.Request(method=method,
                                        url=request_endpoint,
                                        headers=headers,
                                        json=dict(body_param_list))
    else:
        http_request = requests.Request(method=method,
                                        url=request_endpoint,
                                        headers=headers)

    response: requests.Response = requests.session().send(http_request.prepare())

    log.debug(f"HTTP Status Code: {response.status_code}")
    # no support for non-JSON output
    click.echo(JsonFormatter(response.json()).format_with_double_quotes())



#!/usr/bin/env python3
"""
Module with throttling logic.
"""
from dataclasses import dataclass
from typing import Callable, Any, List, Optional
from time import sleep

import click


@dataclass
class ApiRequestThrottler:
    """
    An object that manages a list of api requests.
    """

    def __init__(self, api_requests: List[Callable[[], Any]],
                 *, throttle_seconds: float = 2.5,
                 throttle_limit: int = 25):
        self.api_requests = api_requests
        self.total_calls = len(api_requests)
        self.throttle_seconds = throttle_seconds
        """How long to wait between api calls that are throttled."""
        self.throttle_limit = throttle_limit
        """Maximum of calls that you can do without throttling."""

    def release(self):
        """Yields the next api requests ready to be executed.

        :return:
        """
        throttle_enabled: bool = self.exceeds_throttle_limit()

        if throttle_enabled:
            click.echo(f"To prevent overheating Habitica, we'll "
                       f"wait {self.throttle_seconds} seconds every request.",
                       err=True)

        for api_request in self.api_requests:
            yield api_request
            if throttle_enabled:
                self.throttle()

    def execute_all_requests(self) -> None:
        """Execute all API request and discard the return values."""
        for api_request in self.release():
            _ = api_request()

    def throttle(self):
        """Wait before executing the next api request."""
        sleep(self.throttle_seconds)

    def exceeds_throttle_limit(self, *,
                               api_call_times: Optional[int] = None) -> bool:
        """Return True if we need Hopla to go easy on the Habitica API."""
        call_times: int = api_call_times or self.total_calls
        return call_times > self.throttle_limit



from typing import Dict

import pytest

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.zoo.foodmodels import FeedStatus, Food, FoodException, FoodStockpile, \
    FoodStockpileBuilder, InvalidFeedStatus


class TestFeedStatus:
    valid_feed_range = [-1, 0, *range(5, 50)]

    @pytest.mark.parametrize("valid_status", valid_feed_range)
    def test__init__ok(self, valid_status: int):
        status = FeedStatus(valid_status)
        assert int(status) == valid_status

    @pytest.mark.parametrize("invalid_status", [
        *range(-10, -1), *range(1, 5), *range(50, 60)
    ])
    def test__init__fail(self, invalid_status: int):
        with pytest.raises(InvalidFeedStatus) as exec_info:
            FeedStatus(invalid_status)

        er_msg = str(exec_info.value)
        assert f"{invalid_status} is invalid" in er_msg

    @pytest.mark.parametrize("valid_status", valid_feed_range)
    def test__repr__ok(self, valid_status: int):
        result: str = repr(FeedStatus(valid_status))
        assert result == f"FeedStatus({valid_status})"

    @pytest.mark.parametrize(
        "feed_status,expected_percentage",
        [(-1, 100), (0, 0), (5, 10), (7, 14), (48, 96)]
    )
    def test_to_percentage_ok(self, feed_status: int,
                              expected_percentage: int):
        feed_status = FeedStatus(feed_status)

        result_percentage = feed_status.to_percentage()

        assert result_percentage == expected_percentage

    @pytest.mark.parametrize(
        "feed_status,expected_available", [
            (-1, False), (0, False), (5, True), (27, True), (42, True)
        ]
    )
    def test_is_pet_available(self, feed_status: int,
                              expected_available: bool):
        status = FeedStatus(feed_status)

        result: bool = status.is_pet_available()

        assert result == expected_available

    @pytest.mark.parametrize(
        "start_status,expected_food_items", [
            (45, 1),
            (44, 2),
            (40, 2),
            (39, 3),
            (5, 9),
        ]
    )
    def test_required_food_items_to_become_mount_favorite(self,
                                                          start_status: int,
                                                          expected_food_items: int):
        feed_status = FeedStatus(start_status)

        is_favorite = True
        result = feed_status.required_food_items_to_become_mount(is_favorite)

        assert result == expected_food_items

    @pytest.mark.parametrize(
        "start_status,expected_food_items", [
            (49, 1),
            (45, 3),
            (44, 3),
            (40, 5),
            (39, 6),
            (6, 22),
            (5, 23),
        ]
    )
    def test_required_food_items_to_become_mount_not_favorite(self,
                                                              start_status: int,
                                                              expected_food_items: int):
        feed_status = FeedStatus(start_status)

        is_favorite = False
        result = feed_status.required_food_items_to_become_mount(is_favorite)

        assert result == expected_food_items


class TestFeedStatusHash:
    _feed_statuses = [
        FeedStatus(-1), FeedStatus(), FeedStatus(5),
        FeedStatus(20), FeedStatus(35), FeedStatus(49)
    ]

    @pytest.mark.parametrize("feed_status", _feed_statuses)
    def test__hash__same_object(self, feed_status: FeedStatus):
        # A minimal requirement for __hash__ is that if 2 FeedStatus objects are
        # identical, they MUST have the same hash.
        assert hash(feed_status) == hash(feed_status)

    @pytest.mark.parametrize(
        "feed_status,equal_feed_status",
        list(zip(_feed_statuses + [FeedStatus()],
                 _feed_statuses + [FeedStatus(5)]))
    )
    def test__hash__same_valued_object(self,
                                       feed_status: FeedStatus,
                                       equal_feed_status: FeedStatus):
        # A minimal requirement for __hash__ if 2 FeedStatus objects
        # are equal (__eq__) , they MUST have the same hash.
        assert feed_status == equal_feed_status
        assert hash(feed_status) == hash(equal_feed_status)


class TestFood:

    @pytest.mark.parametrize("food_name", [
        "RottenMeat", "CottonCandyBlue", "Chocolate", "Fish",
        "Saddle", "Cake_Skeleton", "Pie_White", "Candy_Zombie"
    ])
    def test___repr__(self, food_name: str):
        food = Food(food_name)
        result = str(food)

        assert result == f"Food({food_name})"

    @pytest.mark.parametrize("food_name,is_rare_expected", [
        ("RottenMeat", False),
        ("CottonCandyPink", False),
        ("Saddle", True),
        ("Cake_Skeleton", True),
        ("Pie_White", True),
        ("Candy_Zombie", True)
    ])
    def test_is_rare(self, food_name: str,
                     is_rare_expected: bool):
        food = Food(food_name)

        result_is_rare: bool = food.is_rare_food_item()

        assert result_is_rare == is_rare_expected


class TestFoodStockpileBuilder:

    def test_build_empty(self, empty_stockpile_dict: Dict[str, int]):
        result_stockpile: FoodStockpile = FoodStockpileBuilder().build()

        assert result_stockpile.as_dict() == empty_stockpile_dict

    def test_build(self, empty_stockpile_dict: Dict[str, int]):
        drop_food = {"RottenMeat": 201, "CottonCandyBlue": 71, "Chocolate": 1}
        food_dict = {"items": {"food": {
            **drop_food, "Saddle": 1, "Cake_Skeleton": 1, "Pie_White": 2, "Candy_Zombie": 1}
        }}
        user = HabiticaUser(user_dict=food_dict)

        result_stockpile = FoodStockpileBuilder().user(user).build()

        expected_food = empty_stockpile_dict
        expected_food.update(drop_food)

        assert result_stockpile.as_dict() == expected_food

    def test___repr__(self, empty_stockpile_dict: Dict[str, int]):
        drop_food = {"RottenMeat": 201}
        food_dict = {"items": {"food": {**drop_food, "Saddle": 11}}}
        user = HabiticaUser(user_dict=food_dict)

        stockpile = FoodStockpileBuilder().user(user)
        result = str(stockpile)

        expected_food = empty_stockpile_dict
        expected_food.update(drop_food)
        expected_repr = f"FoodStockpileBuilder({expected_food})"
        assert result == expected_repr

    def test_empty_stockpile(self):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        assert all(amount == 0 for amount in stockpile.as_dict().values())


@pytest.fixture
def empty_stockpile_dict() -> Dict[str, int]:
    return {'Meat': 0, 'Milk': 0, 'Potatoe': 0, 'Strawberry': 0,
            'Chocolate': 0, 'Fish': 0, 'RottenMeat': 0,
            'CottonCandyPink': 0, 'CottonCandyBlue': 0, 'Honey': 0}


class TestFoodStockpile:
    def test_add_food_ok(self, empty_stockpile_dict: Dict[str, int]):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        food1, amount1 = "Meat", 5
        food2, amount2 = "Honey", 1
        stockpile.add_food(food1, n=amount1)
        stockpile.add_food(food2, n=amount2)

        expected_food = empty_stockpile_dict
        expected_food.update({food1: amount1, food2: amount2})
        assert stockpile.as_dict() == expected_food
        assert str(stockpile) == f"FoodStockpile({expected_food})"

    def test_add_food_supply_too_small_fail(self):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        food_name = "CottonCandyBlue"
        with pytest.raises(FoodException) as execinfo:
            # can't subtract anything anything from empty stockpile
            stockpile.add_food(food_name, n=-1)

        msg = str(execinfo.value)
        expected_msg = (f"Insufficient food: Cannot remove n=-1 of food from the stockpile\n"
                        f"The current quantity of {food_name} is 0.")

        expected_food = Food(food_name)

        assert msg == expected_msg
        assert execinfo.value.food == expected_food

    def test_add_rare_food_item_fail(self):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        food_name = "Saddle"
        with pytest.raises(FoodException) as execinfo:
            # Saddles are not supported by our stockpile
            stockpile.add_food(food_name, n=1)

        msg = str(execinfo.value)
        expected_msg = f"Not Supported: {food_name=} is not supported."
        expected_food = Food(food_name)

        assert msg == expected_msg
        assert execinfo.value.food == expected_food

    def test_add_food_dict(self):
        food_dict = {"Meat": 9, "CottonCandyBlue": 8, "Honey": 7}
        food_dict2 = {"Meat": -3, "Fish": 1}

        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        stockpile.add_food_dict(food_dict)
        stockpile.add_food_dict(food_dict2)

        expected_stockpile = FoodStockpile(
            {'Chocolate': 0, 'CottonCandyBlue': 8, 'CottonCandyPink': 0, 'Fish': 1, 'Honey': 7,
             'Meat': 6, 'Milk': 0, 'Potatoe': 0, 'RottenMeat': 0, 'Strawberry': 0})
        assert stockpile == expected_stockpile

    def test_get_most_abundant_food(self):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        most_abundant_food, most_abundant_frequency = "Strawberry", 10
        stockpile.add_food(most_abundant_food, n=most_abundant_frequency)
        stockpile.add_food("Meat", n=most_abundant_frequency - 2)
        stockpile.add_food("CottonCandyBlue", n=most_abundant_frequency - 5)

        result_food: str = stockpile.get_most_abundant_food()

        assert result_food == most_abundant_food

    @pytest.mark.parametrize(
        "food_name,n", [("Meat", 0), ("Milk", 8), ("Honey", 10), ("Strawberry", 11)]
    )
    def test_has_sufficient(self, food_name: str, n: int):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        start_quantity = 10
        stockpile.add_food(food_name, n=start_quantity)

        result_sufficient: bool = stockpile.has_sufficient(food_name, n=n)

        # If the start quantity is larger than the required n items, then you have
        #  sufficient number of this food item.
        expected_sufficient = start_quantity >= n
        assert result_sufficient == expected_sufficient

    def test_has_sufficient_abundant_food(self):
        stockpile: FoodStockpile = FoodStockpileBuilder.empty_stockpile()

        abundant_food, abundant_food_quantity = "Meat", 12
        other_food = "Strawberry"
        stockpile.add_food(abundant_food, n=abundant_food_quantity)
        stockpile.add_food(other_food, n=abundant_food_quantity - 10)

        assert stockpile.has_sufficient_abundant_food(n=0)
        assert stockpile.has_sufficient_abundant_food(n=abundant_food_quantity - 1)
        assert stockpile.has_sufficient_abundant_food(n=abundant_food_quantity)
        assert stockpile.has_sufficient_abundant_food(n=abundant_food_quantity + 1) is False



"""
The module with CLI code that handles the `hopla version` command.
"""

import logging

import click

from hopla.hoplalib.hoplaversion import HoplaVersion

log = logging.getLogger()


@click.command()
def version():
    """Print the Hopla version string.

    \f
    :return:
    """
    log.debug("hopla version")
    click.echo(HoplaVersion().semantic_version())



"""
A module for performing feeding Pet HTTP requests.
"""
from typing import NoReturn, Optional, Union

import requests

from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.zoo.feed_clickhelper import get_feed_data_or_exit
from hopla.hoplalib.zoo.zoofeed_algorithms import FeedPlanItem


class FeedPostRequester(HabiticaRequest):
    """
    The FeedPostRequester sends a post request to feed a pet.

    Note: this API endpoint expects query params instead
    of a request body (even though it is a HTTP POST).

    [APIDOCS](https://habitica.com/apidoc/#api-User-UserFeed)
    """
    _DEFAULT_FOOD_AMOUNT = 1

    def __init__(self, *,
                 pet_name: str,
                 food_name: str,
                 food_amount: Optional[int] = _DEFAULT_FOOD_AMOUNT):
        self.pet_name = pet_name
        self.food_name = food_name
        self.query_params = {
            "amount": food_amount or FeedPostRequester._DEFAULT_FOOD_AMOUNT
        }

    @property
    def path(self) -> str:
        """Return the URL used to feed a pet"""
        return f"/user/feed/{self.pet_name}/{self.food_name}"

    @property
    def feed_pet_food_url(self) -> str:
        """Return the url to feed a pet"""
        return UrlBuilder(path_extension=self.path).url

    def post_feed_request(self) -> requests.Response:
        """Performs the feed pet post requests and return the response"""
        return requests.post(url=self.feed_pet_food_url, headers=self.default_headers,
                             params=self.query_params)

    def post_feed_request_get_data_or_exit(self) -> Union[NoReturn, dict]:
        """
        Performs the feed pet post requests and return
        the feed response if successful. Else exit

        :return:
        """
        response: requests.Response = self.post_feed_request()
        return get_feed_data_or_exit(response)

    @classmethod
    def build_from(cls, feed_item: FeedPlanItem) -> "FeedPostRequester":
        """Create a request from a feed plan item."""
        return FeedPostRequester(
            pet_name=feed_item.pet_name,
            food_name=feed_item.food_name,
            food_amount=feed_item.times
        )



#!/usr/bin/env python3
import random
from typing import List

import click
import pytest

from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.hatchery.hatchdata import HatchPotionData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion, HatchPotionCollection, \
    HatchPotionException

_SAMPLE_SIZE = 10


class TestHatchPotion:
    def test__init__invalid_name_fail(self):
        name = "InvalidName"
        with pytest.raises(HatchPotionException) as exec_info:
            HatchPotion(name, quantity=1)

        assert str(exec_info.value).startswith(f"{name} is not a valid hatching potion name.")
        assert exec_info.errisinstance((YouFoundABugRewardError, click.ClickException))

    @pytest.mark.parametrize(
        "potion_name,quantity",
        list(zip(random.sample(HatchPotionData.hatch_potion_names, k=_SAMPLE_SIZE),
                 range(-_SAMPLE_SIZE, 0)))
    )
    def test__init__invalid_quantity_fail(self, potion_name: str, quantity: int):
        with pytest.raises(HatchPotionException) as exec_info:
            HatchPotion(potion_name, quantity=quantity)

        assert str(exec_info.value).startswith(f"{quantity} is below 0.")
        assert exec_info.errisinstance((YouFoundABugRewardError, click.ClickException))

    @pytest.mark.parametrize(
        "potion_name,quantity",
        list(zip(random.sample(HatchPotionData.hatch_potion_names, k=_SAMPLE_SIZE),
                 range(0, _SAMPLE_SIZE)))
    )
    def test__repr__ok(self, potion_name: str, quantity: int):
        potion = HatchPotion(potion_name, quantity=quantity)

        result: str = repr(potion)

        assert result == f"HatchPotion({potion_name}: {quantity})"

    def test__eq__(self):
        assert HatchPotion("Red") == HatchPotion("Red")
        assert HatchPotion("Shimmer", quantity=1) == HatchPotion("Shimmer")
        assert HatchPotion("Silver") != HatchPotion("Silver", quantity=2)
        assert HatchPotion("Watery") != HatchPotion("Glow")

    @pytest.mark.parametrize("potion_name,quantity", [
        ("Base", 10),
        ("CottonCandyBlue", 1),
        ("Golden", 0),
    ])
    def test_is_standard_potion(self, potion_name: str, quantity: int):
        potion = HatchPotion(potion_name, quantity=quantity)

        assert potion.is_standard_hatch_potion() is True
        assert potion.is_magic_hatch_potion() is False
        assert potion.is_wacky_hatch_potion() is False

    @pytest.mark.parametrize("potion_name,quantity", [
        ("BirchBark", 10),
        ("Windup", 1),
        ("Vampire", 0),
        ("Ruby", 9),
        ("Amber", 69),
        ("MossyStone", 42),
        ("SolarSystem", 9001),
    ])
    def test_is_magic_potion(self, potion_name: str, quantity: int):
        potion = HatchPotion(potion_name, quantity=quantity)

        assert potion.is_standard_hatch_potion() is False
        assert potion.is_magic_hatch_potion() is True
        assert potion.is_wacky_hatch_potion() is False

    @pytest.mark.parametrize("potion_name,quantity", [
        ("Veggie", 10),
        ("Dessert", 0),
    ])
    def test_is_wacky_hatch_potion(self, potion_name: str, quantity: int):
        potion = HatchPotion(potion_name, quantity=quantity)

        assert potion.is_standard_hatch_potion() is False
        assert potion.is_magic_hatch_potion() is False
        assert potion.is_wacky_hatch_potion() is True


class TestHatchPotionCollection:
    def test__init__empty_ok(self):
        collection = HatchPotionCollection()
        assert collection == HatchPotionCollection({})
        assert len(collection) == 0

    def test__init__ok(self):
        potion_dict = {"Base": 0, "Moonglow": 42, "Sunset": 2}

        collection = HatchPotionCollection(potion_dict)

        assert collection["Base"] == HatchPotion("Base", quantity=0)
        assert collection["Moonglow"] == HatchPotion("Moonglow", quantity=42)
        assert collection["Sunset"] == HatchPotion("Sunset", quantity=2)

    def test__eq__ok(self):
        left = HatchPotionCollection({"Frost": 1, "Glow": 1})
        right = HatchPotionCollection({"Glow": 1, "Frost": 2})
        assert left != right

        assert HatchPotionCollection() == HatchPotionCollection()
        assert HatchPotionCollection({"StarryNight": 1}) != HatchPotionCollection()
        assert HatchPotionCollection({"Windup": 2}) == HatchPotionCollection({"Windup": 2})
        assert HatchPotionCollection({"Frost": 1}) != HatchPotionCollection({"Frost": 2})

    def test__iter__ok(self):
        collection = HatchPotionCollection({"Base": 1, "Moonglow": 42, "Sunset": 2})

        iterator = iter(collection)

        assert next(iterator) == "Base"
        assert next(iterator) == "Moonglow"
        assert next(iterator) == "Sunset"
        with pytest.raises(StopIteration):
            next(iterator)

    def test__getitem__ok(self):
        collection = HatchPotionCollection({"Base": 1, "Moonglow": 42, "Sunset": 0})

        assert collection["Base"] == HatchPotion("Base", quantity=1)
        assert collection["Moonglow"] == HatchPotion("Moonglow", quantity=42)
        assert collection["Sunset"] == HatchPotion("Sunset", quantity=0)

    def test_values_ok(self):
        potion1, quantity1 = "Dessert", 10
        potion2, quantity2 = "MossyStone", 1
        potion3, quantity3 = "StainedGlass", 2
        collection = HatchPotionCollection({
            potion1: quantity1, potion2: quantity2, potion3: quantity3
        })

        generator = collection.values()
        assert next(generator) == HatchPotion(potion1, quantity=quantity1)
        assert next(generator) == HatchPotion(potion2, quantity=quantity2)
        assert next(generator) == HatchPotion(potion3, quantity=quantity3)
        with pytest.raises(StopIteration):
            _ = next(generator)

    def test_values_as_list_ok(self):
        potion1, quantity1 = "Golden", 1
        potion2, quantity2 = "Sunshine", 41
        potion3, quantity3 = "Vampire", 3
        collection = HatchPotionCollection({
            potion1: quantity1, potion2: quantity2, potion3: quantity3
        })

        result: List[HatchPotion] = list(collection.values())

        expected: List[HatchPotion] = [
            HatchPotion(potion1, quantity=quantity1),
            HatchPotion(potion2, quantity=quantity2),
            HatchPotion(potion3, quantity=quantity3)
        ]
        assert result == expected

    def test_remove_hatch_potion_ok(self):
        potion1_quantity = 3
        potion2_quantity = 42
        potion3_name, potion3_quantity = "Sunset", 1
        collection = HatchPotionCollection({
            "Base": potion1_quantity,
            "Moonglow": potion2_quantity,
            potion3_name: potion3_quantity
        })

        collection.remove_hatch_potion(HatchPotion("Base"))
        collection.remove_hatch_potion(HatchPotion("Moonglow"))
        collection.remove_hatch_potion(HatchPotion(potion3_name))

        assert collection["Base"] == HatchPotion("Base",
                                                 quantity=potion1_quantity - 1)
        assert collection["Moonglow"] == HatchPotion("Moonglow",
                                                     quantity=potion2_quantity - 1)
        assert collection[potion3_name] == HatchPotion(potion3_name,
                                                       quantity=potion3_quantity - 1)

    def test_remove_hatch_potion_not_available_faile(self):
        collection = HatchPotionCollection({"Base": 1})

        not_found_potion_name = "Moonglow"
        with pytest.raises(HatchPotionException) as exec_info:
            collection.remove_hatch_potion(HatchPotion(not_found_potion_name))

        expected_msg = f"{not_found_potion_name} was not in the collection "
        assert str(exec_info.value).startswith(expected_msg)



"""
A helper module with Pet Data.

Later on this module may be refactored to dynamically retrieve pet
data from the API. The retrieved data from the API can then be
cached locally. We then need to refresh our cache only when the
appVersion of the Habitica API increases. However, this approach
is overkill for now.
"""


class PetData:
    """ Helper class with valid Habitica pets.

    For more info:
    * hopla api content | jq .petInfo
    * hopla api content | jq .mountInfo

    And also interesting:
    * hopla api content | jq .premiumPets
    * hopla api content | jq .pets
    * hopla api content | jq .questPets
    * hopla api content | jq .specialPets
    * hopla api content | jq .wackyPets
    """

    generation1_pet_names = [
        "Wolf-Base", "Wolf-White", "Wolf-Desert", "Wolf-Red", "Wolf-Shade", "Wolf-Skeleton",
        "Wolf-Zombie", "Wolf-CottonCandyPink", "Wolf-CottonCandyBlue", "Wolf-Golden",
        "TigerCub-Base", "TigerCub-White", "TigerCub-Desert", "TigerCub-Red", "TigerCub-Shade",
        "TigerCub-Skeleton", "TigerCub-Zombie", "TigerCub-CottonCandyPink",
        "TigerCub-CottonCandyBlue", "TigerCub-Golden", "PandaCub-Base", "PandaCub-White",
        "PandaCub-Desert", "PandaCub-Red", "PandaCub-Shade", "PandaCub-Skeleton",
        "PandaCub-Zombie", "PandaCub-CottonCandyPink", "PandaCub-CottonCandyBlue",
        "PandaCub-Golden", "LionCub-Base", "LionCub-White", "LionCub-Desert", "LionCub-Red",
        "LionCub-Shade", "LionCub-Skeleton", "LionCub-Zombie", "LionCub-CottonCandyPink",
        "LionCub-CottonCandyBlue", "LionCub-Golden", "Fox-Base", "Fox-White", "Fox-Desert",
        "Fox-Red", "Fox-Shade", "Fox-Skeleton", "Fox-Zombie", "Fox-CottonCandyPink",
        "Fox-CottonCandyBlue", "Fox-Golden", "FlyingPig-Base", "FlyingPig-White",
        "FlyingPig-Desert", "FlyingPig-Red", "FlyingPig-Shade", "FlyingPig-Skeleton",
        "FlyingPig-Zombie", "FlyingPig-CottonCandyPink", "FlyingPig-CottonCandyBlue",
        "FlyingPig-Golden", "Dragon-Base", "Dragon-White", "Dragon-Desert", "Dragon-Red",
        "Dragon-Shade", "Dragon-Skeleton", "Dragon-Zombie", "Dragon-CottonCandyPink",
        "Dragon-CottonCandyBlue", "Dragon-Golden", "Cactus-Base", "Cactus-White", "Cactus-Desert",
        "Cactus-Red", "Cactus-Shade", "Cactus-Skeleton", "Cactus-Zombie", "Cactus-CottonCandyPink",
        "Cactus-CottonCandyBlue", "Cactus-Golden", "BearCub-Base", "BearCub-White",
        "BearCub-Desert", "BearCub-Red", "BearCub-Shade", "BearCub-Skeleton", "BearCub-Zombie",
        "BearCub-CottonCandyPink", "BearCub-CottonCandyBlue", "BearCub-Golden",
    ]
    """ https://habitica.fandom.com/wiki/Pets#Generation_1_Pets """

    # The Magic potion pets are known as "premium" by the API
    # <https://habitica.fandom.com/wiki/Pets#Magic_Potion_Pets>
    # hopla api content | jq '[.petInfo[] | select(.type=="premium")]'
    magic_potion_pet_names = [
        "Wolf-RoyalPurple", "Wolf-Cupid", "Wolf-Shimmer", "Wolf-Fairy", "Wolf-Floral",
        "Wolf-Aquatic", "Wolf-Ember", "Wolf-Thunderstorm", "Wolf-Spooky", "Wolf-Ghost",
        "Wolf-Holly", "Wolf-Peppermint", "Wolf-StarryNight", "Wolf-Rainbow", "Wolf-Glass",
        "Wolf-Glow", "Wolf-Frost", "Wolf-IcySnow", "Wolf-RoseQuartz", "Wolf-Celestial",
        "Wolf-Sunshine", "Wolf-Bronze", "Wolf-Watery", "Wolf-Silver", "Wolf-Shadow", "Wolf-Amber",
        "Wolf-Aurora", "Wolf-Ruby", "Wolf-BirchBark", "Wolf-Fluorite", "Wolf-SandSculpture",
        "Wolf-Windup", "Wolf-Turquoise", "Wolf-Vampire", "Wolf-AutumnLeaf", "Wolf-BlackPearl",
        "Wolf-StainedGlass", "Wolf-PolkaDot", "Wolf-MossyStone", "Wolf-Sunset", "Wolf-Moonglow",
        "Wolf-SolarSystem", "TigerCub-RoyalPurple", "TigerCub-Cupid", "TigerCub-Shimmer",
        "TigerCub-Fairy", "TigerCub-Floral", "TigerCub-Aquatic", "TigerCub-Ember",
        "TigerCub-Thunderstorm", "TigerCub-Spooky", "TigerCub-Ghost", "TigerCub-Holly",
        "TigerCub-Peppermint", "TigerCub-StarryNight", "TigerCub-Rainbow", "TigerCub-Glass",
        "TigerCub-Glow", "TigerCub-Frost", "TigerCub-IcySnow", "TigerCub-RoseQuartz",
        "TigerCub-Celestial", "TigerCub-Sunshine", "TigerCub-Bronze", "TigerCub-Watery",
        "TigerCub-Silver", "TigerCub-Shadow", "TigerCub-Amber", "TigerCub-Aurora", "TigerCub-Ruby",
        "TigerCub-BirchBark", "TigerCub-Fluorite", "TigerCub-SandSculpture", "TigerCub-Windup",
        "TigerCub-Turquoise", "TigerCub-Vampire", "TigerCub-AutumnLeaf", "TigerCub-BlackPearl",
        "TigerCub-StainedGlass", "TigerCub-PolkaDot", "TigerCub-MossyStone", "TigerCub-Sunset",
        "TigerCub-Moonglow", "TigerCub-SolarSystem", "PandaCub-RoyalPurple", "PandaCub-Cupid",
        "PandaCub-Shimmer", "PandaCub-Fairy", "PandaCub-Floral", "PandaCub-Aquatic",
        "PandaCub-Ember", "PandaCub-Thunderstorm", "PandaCub-Spooky", "PandaCub-Ghost",
        "PandaCub-Holly", "PandaCub-Peppermint", "PandaCub-StarryNight", "PandaCub-Rainbow",
        "PandaCub-Glass", "PandaCub-Glow", "PandaCub-Frost", "PandaCub-IcySnow",
        "PandaCub-RoseQuartz", "PandaCub-Celestial", "PandaCub-Sunshine", "PandaCub-Bronze",
        "PandaCub-Watery", "PandaCub-Silver", "PandaCub-Shadow", "PandaCub-Amber",
        "PandaCub-Aurora", "PandaCub-Ruby", "PandaCub-BirchBark", "PandaCub-Fluorite",
        "PandaCub-SandSculpture", "PandaCub-Windup", "PandaCub-Turquoise", "PandaCub-Vampire",
        "PandaCub-AutumnLeaf", "PandaCub-BlackPearl", "PandaCub-StainedGlass", "PandaCub-PolkaDot",
        "PandaCub-MossyStone", "PandaCub-Sunset", "PandaCub-Moonglow", "PandaCub-SolarSystem",
        "LionCub-RoyalPurple", "LionCub-Cupid", "LionCub-Shimmer", "LionCub-Fairy",
        "LionCub-Floral", "LionCub-Aquatic", "LionCub-Ember", "LionCub-Thunderstorm",
        "LionCub-Spooky", "LionCub-Ghost", "LionCub-Holly", "LionCub-Peppermint",
        "LionCub-StarryNight", "LionCub-Rainbow", "LionCub-Glass", "LionCub-Glow", "LionCub-Frost",
        "LionCub-IcySnow", "LionCub-RoseQuartz", "LionCub-Celestial", "LionCub-Sunshine",
        "LionCub-Bronze", "LionCub-Watery", "LionCub-Silver", "LionCub-Shadow", "LionCub-Amber",
        "LionCub-Aurora", "LionCub-Ruby", "LionCub-BirchBark", "LionCub-Fluorite",
        "LionCub-SandSculpture", "LionCub-Windup", "LionCub-Turquoise", "LionCub-Vampire",
        "LionCub-AutumnLeaf", "LionCub-BlackPearl", "LionCub-StainedGlass", "LionCub-PolkaDot",
        "LionCub-MossyStone", "LionCub-Sunset", "LionCub-Moonglow", "LionCub-SolarSystem",
        "Fox-RoyalPurple", "Fox-Cupid", "Fox-Shimmer", "Fox-Fairy", "Fox-Floral", "Fox-Aquatic",
        "Fox-Ember", "Fox-Thunderstorm", "Fox-Spooky", "Fox-Ghost", "Fox-Holly", "Fox-Peppermint",
        "Fox-StarryNight", "Fox-Rainbow", "Fox-Glass", "Fox-Glow", "Fox-Frost", "Fox-IcySnow",
        "Fox-RoseQuartz", "Fox-Celestial", "Fox-Sunshine", "Fox-Bronze", "Fox-Watery",
        "Fox-Silver", "Fox-Shadow", "Fox-Amber", "Fox-Aurora", "Fox-Ruby", "Fox-BirchBark",
        "Fox-Fluorite", "Fox-SandSculpture", "Fox-Windup", "Fox-Turquoise", "Fox-Vampire",
        "Fox-AutumnLeaf", "Fox-BlackPearl", "Fox-StainedGlass", "Fox-PolkaDot", "Fox-MossyStone",
        "Fox-Sunset", "Fox-Moonglow", "Fox-SolarSystem", "FlyingPig-RoyalPurple",
        "FlyingPig-Cupid", "FlyingPig-Shimmer", "FlyingPig-Fairy", "FlyingPig-Floral",
        "FlyingPig-Aquatic", "FlyingPig-Ember", "FlyingPig-Thunderstorm", "FlyingPig-Spooky",
        "FlyingPig-Ghost", "FlyingPig-Holly", "FlyingPig-Peppermint", "FlyingPig-StarryNight",
        "FlyingPig-Rainbow", "FlyingPig-Glass", "FlyingPig-Glow", "FlyingPig-Frost",
        "FlyingPig-IcySnow", "FlyingPig-RoseQuartz", "FlyingPig-Celestial", "FlyingPig-Sunshine",
        "FlyingPig-Bronze", "FlyingPig-Watery", "FlyingPig-Silver", "FlyingPig-Shadow",
        "FlyingPig-Amber", "FlyingPig-Aurora", "FlyingPig-Ruby", "FlyingPig-BirchBark",
        "FlyingPig-Fluorite", "FlyingPig-SandSculpture", "FlyingPig-Windup", "FlyingPig-Turquoise",
        "FlyingPig-Vampire", "FlyingPig-AutumnLeaf", "FlyingPig-BlackPearl",
        "FlyingPig-StainedGlass", "FlyingPig-PolkaDot", "FlyingPig-MossyStone", "FlyingPig-Sunset",
        "FlyingPig-Moonglow", "FlyingPig-SolarSystem", "Dragon-RoyalPurple", "Dragon-Cupid",
        "Dragon-Shimmer", "Dragon-Fairy", "Dragon-Floral", "Dragon-Aquatic", "Dragon-Ember",
        "Dragon-Thunderstorm", "Dragon-Spooky", "Dragon-Ghost", "Dragon-Holly",
        "Dragon-Peppermint", "Dragon-StarryNight", "Dragon-Rainbow", "Dragon-Glass", "Dragon-Glow",
        "Dragon-Frost", "Dragon-IcySnow", "Dragon-RoseQuartz", "Dragon-Celestial",
        "Dragon-Sunshine", "Dragon-Bronze", "Dragon-Watery", "Dragon-Silver", "Dragon-Shadow",
        "Dragon-Amber", "Dragon-Aurora", "Dragon-Ruby", "Dragon-BirchBark", "Dragon-Fluorite",
        "Dragon-SandSculpture", "Dragon-Windup", "Dragon-Turquoise", "Dragon-Vampire",
        "Dragon-AutumnLeaf", "Dragon-BlackPearl", "Dragon-StainedGlass", "Dragon-PolkaDot",
        "Dragon-MossyStone", "Dragon-Sunset", "Dragon-Moonglow", "Dragon-SolarSystem",
        "Cactus-RoyalPurple", "Cactus-Cupid", "Cactus-Shimmer", "Cactus-Fairy", "Cactus-Floral",
        "Cactus-Aquatic", "Cactus-Ember", "Cactus-Thunderstorm", "Cactus-Spooky", "Cactus-Ghost",
        "Cactus-Holly", "Cactus-Peppermint", "Cactus-StarryNight", "Cactus-Rainbow",
        "Cactus-Glass", "Cactus-Glow", "Cactus-Frost", "Cactus-IcySnow", "Cactus-RoseQuartz",
        "Cactus-Celestial", "Cactus-Sunshine", "Cactus-Bronze", "Cactus-Watery", "Cactus-Silver",
        "Cactus-Shadow", "Cactus-Amber", "Cactus-Aurora", "Cactus-Ruby", "Cactus-BirchBark",
        "Cactus-Fluorite", "Cactus-SandSculpture", "Cactus-Windup", "Cactus-Turquoise",
        "Cactus-Vampire", "Cactus-AutumnLeaf", "Cactus-BlackPearl", "Cactus-StainedGlass",
        "Cactus-PolkaDot", "Cactus-MossyStone", "Cactus-Sunset", "Cactus-Moonglow",
        "Cactus-SolarSystem", "BearCub-RoyalPurple", "BearCub-Cupid", "BearCub-Shimmer",
        "BearCub-Fairy", "BearCub-Floral", "BearCub-Aquatic", "BearCub-Ember",
        "BearCub-Thunderstorm", "BearCub-Spooky", "BearCub-Ghost", "BearCub-Holly",
        "BearCub-Peppermint", "BearCub-StarryNight", "BearCub-Rainbow", "BearCub-Glass",
        "BearCub-Glow", "BearCub-Frost", "BearCub-IcySnow", "BearCub-RoseQuartz",
        "BearCub-Celestial", "BearCub-Sunshine", "BearCub-Bronze", "BearCub-Watery",
        "BearCub-Silver", "BearCub-Shadow", "BearCub-Amber", "BearCub-Aurora", "BearCub-Ruby",
        "BearCub-BirchBark", "BearCub-Fluorite", "BearCub-SandSculpture", "BearCub-Windup",
        "BearCub-Turquoise", "BearCub-Vampire", "BearCub-AutumnLeaf", "BearCub-BlackPearl",
        "BearCub-StainedGlass", "BearCub-PolkaDot", "BearCub-MossyStone", "BearCub-Sunset",
        "BearCub-Moonglow", "BearCub-SolarSystem",
    ]

    # Quest pets: @see:
    # <https://habitica.fandom.com/wiki/Pets#Quest_Pets>
    # <https://habitica.fandom.com/wiki/Pets#Magic_Potion_Pets>
    # hopla api content | jq '[.petInfo[] | select(.type=="quest")]'
    quest_pet_names = [
        "Gryphon-Base", "Gryphon-White", "Gryphon-Desert", "Gryphon-Red", "Gryphon-Shade",
        "Gryphon-Skeleton", "Gryphon-Zombie", "Gryphon-CottonCandyPink", "Gryphon-CottonCandyBlue",
        "Gryphon-Golden", "Hedgehog-Base", "Hedgehog-White", "Hedgehog-Desert", "Hedgehog-Red",
        "Hedgehog-Shade", "Hedgehog-Skeleton", "Hedgehog-Zombie", "Hedgehog-CottonCandyPink",
        "Hedgehog-CottonCandyBlue", "Hedgehog-Golden", "Deer-Base", "Deer-White", "Deer-Desert",
        "Deer-Red", "Deer-Shade", "Deer-Skeleton", "Deer-Zombie", "Deer-CottonCandyPink",
        "Deer-CottonCandyBlue", "Deer-Golden", "Egg-Base", "Egg-White", "Egg-Desert", "Egg-Red",
        "Egg-Shade", "Egg-Skeleton", "Egg-Zombie", "Egg-CottonCandyPink", "Egg-CottonCandyBlue",
        "Egg-Golden", "Rat-Base", "Rat-White", "Rat-Desert", "Rat-Red", "Rat-Shade",
        "Rat-Skeleton", "Rat-Zombie", "Rat-CottonCandyPink", "Rat-CottonCandyBlue", "Rat-Golden",
        "Octopus-Base", "Octopus-White", "Octopus-Desert", "Octopus-Red", "Octopus-Shade",
        "Octopus-Skeleton", "Octopus-Zombie", "Octopus-CottonCandyPink", "Octopus-CottonCandyBlue",
        "Octopus-Golden", "Seahorse-Base", "Seahorse-White", "Seahorse-Desert", "Seahorse-Red",
        "Seahorse-Shade", "Seahorse-Skeleton", "Seahorse-Zombie", "Seahorse-CottonCandyPink",
        "Seahorse-CottonCandyBlue", "Seahorse-Golden", "Parrot-Base", "Parrot-White",
        "Parrot-Desert", "Parrot-Red", "Parrot-Shade", "Parrot-Skeleton", "Parrot-Zombie",
        "Parrot-CottonCandyPink", "Parrot-CottonCandyBlue", "Parrot-Golden", "Rooster-Base",
        "Rooster-White", "Rooster-Desert", "Rooster-Red", "Rooster-Shade", "Rooster-Skeleton",
        "Rooster-Zombie", "Rooster-CottonCandyPink", "Rooster-CottonCandyBlue", "Rooster-Golden",
        "Spider-Base", "Spider-White", "Spider-Desert", "Spider-Red", "Spider-Shade",
        "Spider-Skeleton", "Spider-Zombie", "Spider-CottonCandyPink", "Spider-CottonCandyBlue",
        "Spider-Golden", "Owl-Base", "Owl-White", "Owl-Desert", "Owl-Red", "Owl-Shade",
        "Owl-Skeleton", "Owl-Zombie", "Owl-CottonCandyPink", "Owl-CottonCandyBlue", "Owl-Golden",
        "Penguin-Base", "Penguin-White", "Penguin-Desert", "Penguin-Red", "Penguin-Shade",
        "Penguin-Skeleton", "Penguin-Zombie", "Penguin-CottonCandyPink", "Penguin-CottonCandyBlue",
        "Penguin-Golden", "TRex-Base", "TRex-White", "TRex-Desert", "TRex-Red", "TRex-Shade",
        "TRex-Skeleton", "TRex-Zombie", "TRex-CottonCandyPink", "TRex-CottonCandyBlue",
        "TRex-Golden", "Rock-Base", "Rock-White", "Rock-Desert", "Rock-Red", "Rock-Shade",
        "Rock-Skeleton", "Rock-Zombie", "Rock-CottonCandyPink", "Rock-CottonCandyBlue",
        "Rock-Golden", "Bunny-Base", "Bunny-White", "Bunny-Desert", "Bunny-Red", "Bunny-Shade",
        "Bunny-Skeleton", "Bunny-Zombie", "Bunny-CottonCandyPink", "Bunny-CottonCandyBlue",
        "Bunny-Golden", "Slime-Base", "Slime-White", "Slime-Desert", "Slime-Red", "Slime-Shade",
        "Slime-Skeleton", "Slime-Zombie", "Slime-CottonCandyPink", "Slime-CottonCandyBlue",
        "Slime-Golden", "Sheep-Base", "Sheep-White", "Sheep-Desert", "Sheep-Red", "Sheep-Shade",
        "Sheep-Skeleton", "Sheep-Zombie", "Sheep-CottonCandyPink", "Sheep-CottonCandyBlue",
        "Sheep-Golden", "Cuttlefish-Base", "Cuttlefish-White", "Cuttlefish-Desert",
        "Cuttlefish-Red", "Cuttlefish-Shade", "Cuttlefish-Skeleton", "Cuttlefish-Zombie",
        "Cuttlefish-CottonCandyPink", "Cuttlefish-CottonCandyBlue", "Cuttlefish-Golden",
        "Whale-Base", "Whale-White", "Whale-Desert", "Whale-Red", "Whale-Shade", "Whale-Skeleton",
        "Whale-Zombie", "Whale-CottonCandyPink", "Whale-CottonCandyBlue", "Whale-Golden",
        "Cheetah-Base", "Cheetah-White", "Cheetah-Desert", "Cheetah-Red", "Cheetah-Shade",
        "Cheetah-Skeleton", "Cheetah-Zombie", "Cheetah-CottonCandyPink", "Cheetah-CottonCandyBlue",
        "Cheetah-Golden", "Horse-Base", "Horse-White", "Horse-Desert", "Horse-Red", "Horse-Shade",
        "Horse-Skeleton", "Horse-Zombie", "Horse-CottonCandyPink", "Horse-CottonCandyBlue",
        "Horse-Golden", "Frog-Base", "Frog-White", "Frog-Desert", "Frog-Red", "Frog-Shade",
        "Frog-Skeleton", "Frog-Zombie", "Frog-CottonCandyPink", "Frog-CottonCandyBlue",
        "Frog-Golden", "Snake-Base", "Snake-White", "Snake-Desert", "Snake-Red", "Snake-Shade",
        "Snake-Skeleton", "Snake-Zombie", "Snake-CottonCandyPink", "Snake-CottonCandyBlue",
        "Snake-Golden", "Unicorn-Base", "Unicorn-White", "Unicorn-Desert", "Unicorn-Red",
        "Unicorn-Shade", "Unicorn-Skeleton", "Unicorn-Zombie", "Unicorn-CottonCandyPink",
        "Unicorn-CottonCandyBlue", "Unicorn-Golden", "Sabretooth-Base", "Sabretooth-White",
        "Sabretooth-Desert", "Sabretooth-Red", "Sabretooth-Shade", "Sabretooth-Skeleton",
        "Sabretooth-Zombie", "Sabretooth-CottonCandyPink", "Sabretooth-CottonCandyBlue",
        "Sabretooth-Golden", "Monkey-Base", "Monkey-White", "Monkey-Desert", "Monkey-Red",
        "Monkey-Shade", "Monkey-Skeleton", "Monkey-Zombie", "Monkey-CottonCandyPink",
        "Monkey-CottonCandyBlue", "Monkey-Golden", "Snail-Base", "Snail-White", "Snail-Desert",
        "Snail-Red", "Snail-Shade", "Snail-Skeleton", "Snail-Zombie", "Snail-CottonCandyPink",
        "Snail-CottonCandyBlue", "Snail-Golden", "Falcon-Base", "Falcon-White", "Falcon-Desert",
        "Falcon-Red", "Falcon-Shade", "Falcon-Skeleton", "Falcon-Zombie", "Falcon-CottonCandyPink",
        "Falcon-CottonCandyBlue", "Falcon-Golden", "Treeling-Base", "Treeling-White",
        "Treeling-Desert", "Treeling-Red", "Treeling-Shade", "Treeling-Skeleton",
        "Treeling-Zombie", "Treeling-CottonCandyPink", "Treeling-CottonCandyBlue",
        "Treeling-Golden", "Axolotl-Base", "Axolotl-White", "Axolotl-Desert", "Axolotl-Red",
        "Axolotl-Shade", "Axolotl-Skeleton", "Axolotl-Zombie", "Axolotl-CottonCandyPink",
        "Axolotl-CottonCandyBlue", "Axolotl-Golden", "Turtle-Base", "Turtle-White",
        "Turtle-Desert", "Turtle-Red", "Turtle-Shade", "Turtle-Skeleton", "Turtle-Zombie",
        "Turtle-CottonCandyPink", "Turtle-CottonCandyBlue", "Turtle-Golden", "Armadillo-Base",
        "Armadillo-White", "Armadillo-Desert", "Armadillo-Red", "Armadillo-Shade",
        "Armadillo-Skeleton", "Armadillo-Zombie", "Armadillo-CottonCandyPink",
        "Armadillo-CottonCandyBlue", "Armadillo-Golden", "Cow-Base", "Cow-White", "Cow-Desert",
        "Cow-Red", "Cow-Shade", "Cow-Skeleton", "Cow-Zombie", "Cow-CottonCandyPink",
        "Cow-CottonCandyBlue", "Cow-Golden", "Beetle-Base", "Beetle-White", "Beetle-Desert",
        "Beetle-Red", "Beetle-Shade", "Beetle-Skeleton", "Beetle-Zombie", "Beetle-CottonCandyPink",
        "Beetle-CottonCandyBlue", "Beetle-Golden", "Ferret-Base", "Ferret-White", "Ferret-Desert",
        "Ferret-Red", "Ferret-Shade", "Ferret-Skeleton", "Ferret-Zombie", "Ferret-CottonCandyPink",
        "Ferret-CottonCandyBlue", "Ferret-Golden", "Sloth-Base", "Sloth-White", "Sloth-Desert",
        "Sloth-Red", "Sloth-Shade", "Sloth-Skeleton", "Sloth-Zombie", "Sloth-CottonCandyPink",
        "Sloth-CottonCandyBlue", "Sloth-Golden", "Triceratops-Base", "Triceratops-White",
        "Triceratops-Desert", "Triceratops-Red", "Triceratops-Shade", "Triceratops-Skeleton",
        "Triceratops-Zombie", "Triceratops-CottonCandyPink", "Triceratops-CottonCandyBlue",
        "Triceratops-Golden", "GuineaPig-Base", "GuineaPig-White", "GuineaPig-Desert",
        "GuineaPig-Red", "GuineaPig-Shade", "GuineaPig-Skeleton", "GuineaPig-Zombie",
        "GuineaPig-CottonCandyPink", "GuineaPig-CottonCandyBlue", "GuineaPig-Golden",
        "Peacock-Base", "Peacock-White", "Peacock-Desert", "Peacock-Red", "Peacock-Shade",
        "Peacock-Skeleton", "Peacock-Zombie", "Peacock-CottonCandyPink", "Peacock-CottonCandyBlue",
        "Peacock-Golden", "Butterfly-Base", "Butterfly-White", "Butterfly-Desert", "Butterfly-Red",
        "Butterfly-Shade", "Butterfly-Skeleton", "Butterfly-Zombie", "Butterfly-CottonCandyPink",
        "Butterfly-CottonCandyBlue", "Butterfly-Golden", "Nudibranch-Base", "Nudibranch-White",
        "Nudibranch-Desert", "Nudibranch-Red", "Nudibranch-Shade", "Nudibranch-Skeleton",
        "Nudibranch-Zombie", "Nudibranch-CottonCandyPink", "Nudibranch-CottonCandyBlue",
        "Nudibranch-Golden", "Hippo-Base", "Hippo-White", "Hippo-Desert", "Hippo-Red",
        "Hippo-Shade", "Hippo-Skeleton", "Hippo-Zombie", "Hippo-CottonCandyPink",
        "Hippo-CottonCandyBlue", "Hippo-Golden", "Yarn-Base", "Yarn-White", "Yarn-Desert",
        "Yarn-Red", "Yarn-Shade", "Yarn-Skeleton", "Yarn-Zombie", "Yarn-CottonCandyPink",
        "Yarn-CottonCandyBlue", "Yarn-Golden", "Pterodactyl-Base", "Pterodactyl-White",
        "Pterodactyl-Desert", "Pterodactyl-Red", "Pterodactyl-Shade", "Pterodactyl-Skeleton",
        "Pterodactyl-Zombie", "Pterodactyl-CottonCandyPink", "Pterodactyl-CottonCandyBlue",
        "Pterodactyl-Golden", "Badger-Base", "Badger-White", "Badger-Desert", "Badger-Red",
        "Badger-Shade", "Badger-Skeleton", "Badger-Zombie", "Badger-CottonCandyPink",
        "Badger-CottonCandyBlue", "Badger-Golden", "Squirrel-Base", "Squirrel-White",
        "Squirrel-Desert", "Squirrel-Red", "Squirrel-Shade", "Squirrel-Skeleton",
        "Squirrel-Zombie", "Squirrel-CottonCandyPink", "Squirrel-CottonCandyBlue",
        "Squirrel-Golden", "SeaSerpent-Base", "SeaSerpent-White", "SeaSerpent-Desert",
        "SeaSerpent-Red", "SeaSerpent-Shade", "SeaSerpent-Skeleton", "SeaSerpent-Zombie",
        "SeaSerpent-CottonCandyPink", "SeaSerpent-CottonCandyBlue", "SeaSerpent-Golden",
        "Kangaroo-Base", "Kangaroo-White", "Kangaroo-Desert", "Kangaroo-Red", "Kangaroo-Shade",
        "Kangaroo-Skeleton", "Kangaroo-Zombie", "Kangaroo-CottonCandyPink",
        "Kangaroo-CottonCandyBlue", "Kangaroo-Golden", "Alligator-Base", "Alligator-White",
        "Alligator-Desert", "Alligator-Red", "Alligator-Shade", "Alligator-Skeleton",
        "Alligator-Zombie", "Alligator-CottonCandyPink", "Alligator-CottonCandyBlue",
        "Alligator-Golden", "Velociraptor-Base", "Velociraptor-White", "Velociraptor-Desert",
        "Velociraptor-Red", "Velociraptor-Shade", "Velociraptor-Skeleton", "Velociraptor-Zombie",
        "Velociraptor-CottonCandyPink", "Velociraptor-CottonCandyBlue", "Velociraptor-Golden",
        "Dolphin-Base", "Dolphin-White", "Dolphin-Desert", "Dolphin-Red", "Dolphin-Shade",
        "Dolphin-Skeleton", "Dolphin-Zombie", "Dolphin-CottonCandyPink", "Dolphin-CottonCandyBlue",
        "Dolphin-Golden", "Robot-Base", "Robot-White", "Robot-Desert", "Robot-Red", "Robot-Shade",
        "Robot-Skeleton", "Robot-Zombie", "Robot-CottonCandyPink", "Robot-CottonCandyBlue",
        "Robot-Golden"
    ]

    # hopla api content | jq '[.petInfo[] | select(.type=="wacky")]'
    wacky_pet_names = [
        "Wolf-Veggie", "Wolf-Dessert", "TigerCub-Veggie", "TigerCub-Dessert", "PandaCub-Veggie",
        "PandaCub-Dessert", "LionCub-Veggie", "LionCub-Dessert", "Fox-Veggie", "Fox-Dessert",
        "FlyingPig-Veggie", "FlyingPig-Dessert", "Dragon-Veggie", "Dragon-Dessert",
        "Cactus-Veggie", "Cactus-Dessert", "BearCub-Veggie", "BearCub-Dessert"
    ]

    world_boss_reward_pet_names = [
        "Hippogriff-Hopeful", "MagicalBee-Base", "Phoenix-Base", "Mammoth-Base",
        "MantisShrimp-Base"
    ]
    # Consider refactoring this entire class to contain more semantic information when important.

    event_sequence_pet_names = [
        "Wolf-Veteran", "Turkey-Base", "JackOLantern-Base", "Tiger-Veteran", "Turkey-Gilded",
        "Lion-Veteran", "Gryphon-RoyalPurple", "JackOLantern-Ghost", "Orca-Base", "Bear-Veteran",
        "Fox-Veteran", "JackOLantern-Glow", "JackOLantern-RoyalPurple"
    ]
    """https://habitica.fandom.com/wiki/Event_Item_Sequences"""

    other_pet_names = [
        "Jackalope-RoyalPurple", "BearCub-Polar", "Dragon-Hydra", "Wolf-Cerberus",
        "Gryphon-Gryphatrice", "Aether-Invisible"
    ]

    rare_pet_names = world_boss_reward_pet_names + event_sequence_pet_names + other_pet_names
    """Rare pet names are pets that cannot be hatched.
    @see: <https://habitica.fandom.com/wiki/Pets#Rare_Pets>
    """

    only_1favorite_food_pet_names = generation1_pet_names + quest_pet_names
    """Only quest pets and gen1 pets have only 1 favorite food (anno Sept. 2021)"""

    feedable_pet_names = only_1favorite_food_pet_names + magic_potion_pet_names
    """
    Pets that can be fed are the gen1 pets, quest pets, and magic potion pets. (anno Sept. 2021).
    """

    unfeedable_pet_names = rare_pet_names + wacky_pet_names
    """Only wacky pets and rare pets cannot be fed anno Sept. 2021."""

    pet_names = (generation1_pet_names
                 + magic_potion_pet_names
                 + quest_pet_names
                 + wacky_pet_names
                 + rare_pet_names)
    """ @see: hopla api content | jq '[.petInfo[] | .key]' """



"""
Library code to handle Hopla's authorization, authentication and
identification.
"""

import uuid
import logging
import sys
from configparser import ConfigParser
from pathlib import Path
from hopla.hoplalib.common import get_configuration_dirpath, EnvironmentVariables

log = logging.getLogger()


class AuthorizationFileConstants:
    """Class with authorization and authentication related constants"""

    CONFIG_SECTION_CREDENTIALS = "credentials"
    """only section of the authorization file"""

    CONFIG_KEY_USER_ID = "user_id"
    """user id field in the hopla authorization file """
    CONFIG_KEY_API_TOKEN = "api_token"
    """api token field in the hopla authorization file """


class HoplaAuthFile:
    """
    Class representing the Hopla Auth file. This file keeps the user uuid and user
    api token separate from other hopla configuration.
    """
    def __init__(self):
        self.global_env_var_hopla_auth_file = EnvironmentVariables.HOPLA_AUTH_FILE

    @property
    def file_path(self) -> Path:
        """ Get the file with authorization"""
        if self.global_env_var_hopla_auth_file is not None:
            auth_file = Path(self.global_env_var_hopla_auth_file)
        else:
            auth_file = get_configuration_dirpath() / "authenticate.conf"

        return auth_file.resolve()

    def exists(self) -> bool:
        """Return True if the authentication file exists"""
        return self.file_path.exists() and self.file_path.is_file()

    def create_auth_dir(self):
        """Create the directory that the authorization file is supposed to be in"""
        Path.mkdir(self.file_path.parent, parents=True, exist_ok=True)


class AuthorizationHandler:
    """
    This class *should* only get and set values in the hopla authorization file.
    """

    def __init__(self, *, auth_file: HoplaAuthFile = None):
        self.config_parser = ConfigParser()
        if auth_file is None:
            self.auth_file = HoplaAuthFile()
        else:
            self.auth_file = auth_file

    @property
    def user_id(self):
        """Return the user id to be used in habitica API requests"""
        self._parse()
        return self.config_parser[AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS] \
            .get(AuthorizationFileConstants.CONFIG_KEY_USER_ID)

    @property
    def api_token(self):
        """Return the api token to be used in habitica API requests"""
        self._parse()
        return self.config_parser[AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS] \
            .get(AuthorizationFileConstants.CONFIG_KEY_API_TOKEN)

    def set_hopla_credentials(self, *,
                              user_id: uuid.UUID,
                              api_token: uuid.UUID,
                              overwrite: bool = False):
        """set credentials in the credentials file

        :param user_id:
        :param api_token:
        :param overwrite:
        :return:
        """
        log.debug(f"set_hopla_credentials overwrite={overwrite}")
        if self.auth_file.exists() and overwrite is False:
            log.info(f"Auth file {self.auth_file} not recreated because it already exists")
            return

        self.auth_file.create_auth_dir()
        with open(self.auth_file.file_path, mode="w", encoding="utf-8") as new_auth_file:
            self.config_parser.add_section(
                AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS)
            self.config_parser.set(
                section=AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS,
                option=AuthorizationFileConstants.CONFIG_KEY_USER_ID,
                value=str(user_id)
            )
            self.config_parser.set(
                section=AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS,
                option=AuthorizationFileConstants.CONFIG_KEY_API_TOKEN,
                value=str(api_token)
            )
            self.config_parser.write(new_auth_file)

        assert self.auth_file_is_valid(), f"{self.auth_file} is not valid"

    def auth_file_is_valid(self) -> bool:
        """
        Return True when the authenticate file exists, has a [credentials] section with
        a user id and api token field in it.
        """
        if self.auth_file.exists() is False:
            log.debug(f"{self.auth_file} does not exist")
            return False

        self.config_parser.read(self.auth_file.file_path)
        if self.config_parser.has_section(
                AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS) is False:
            log.debug(f"{self.auth_file} has no "
                      f"{AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS} section")
            return False

        if self._auth_file_has_user_id() is False:
            log.debug(f"{self.auth_file} has no user id")
            return False

        if self._auth_file_has_api_token() is False:
            log.debug(f"{self.auth_file} has no api token")
            return False

        return True

    def _auth_file_has_api_token(self) -> bool:
        """
        Return true if the authenticate file has an api_token value
        inside the [credentials] section.
        """
        return self.config_parser.has_option(
            section=AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS,
            option=AuthorizationFileConstants.CONFIG_KEY_API_TOKEN)

    def _auth_file_has_user_id(self) -> bool:
        """
        Return true if the authenticate file has an user_id
        value inside the [credentials] section.
        """
        return self.config_parser.has_option(
            section=AuthorizationFileConstants.CONFIG_SECTION_CREDENTIALS,
            option=AuthorizationFileConstants.CONFIG_KEY_USER_ID)

    def _parse(self):
        if self.auth_file.exists():
            self.config_parser.read(self.auth_file.file_path)
        if self._auth_file_has_api_token() is False or self._auth_file_has_user_id() is False:
            print("no credentials found")
            print("Please run:")
            print("    hopla authenticate")
            sys.exit(1)



#!/usr/bin/env python3
import random
from typing import List

import click
import pytest

from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.hatchery.eggmodels import Egg, EggCollection, EggException
from hopla.hoplalib.hatchery.hatchdata import EggData, HatchPotionData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion

_SAMPLE_SIZE = 9


class TestEgg:
    def test__init__invalid_name_fail(self):
        name = "InvalidName"
        with pytest.raises(EggException) as exec_info:
            Egg(name)

        assert str(exec_info.value).startswith(f"{name} is not a valid egg name.")
        assert exec_info.errisinstance((YouFoundABugRewardError, click.ClickException))

    @pytest.mark.parametrize(
        "egg_name,quantity",
        list(zip(random.sample(EggData.egg_names, k=_SAMPLE_SIZE),
                 range(-_SAMPLE_SIZE, 0)))
    )
    def test__init__invalid_quantity_fail(self, egg_name: str, quantity: int):
        with pytest.raises(EggException) as exec_info:
            Egg(egg_name, quantity=quantity)

        assert str(exec_info.value).startswith(f"{quantity} is below 0.")
        assert exec_info.errisinstance((YouFoundABugRewardError, click.ClickException))

    @pytest.mark.parametrize("egg1,egg2,expected_equals", [
        (Egg("Horse"), Egg("Horse"), True),
        (Egg("Seahorse", quantity=1), Egg("Seahorse"), True),
        (Egg("Snake", quantity=1), Egg("Snake", quantity=2), False),
        (Egg("Whale"), Egg("Rat"), False),
        (Egg("Rooster", quantity=0), Egg("TRex", quantity=0), False),
        (Egg("Penguin", quantity=1), Egg("Owl", quantity=4), False),
    ])
    def test__eq__(self, egg1: Egg, egg2: Egg, expected_equals: bool):
        assert (egg1 == egg2) is expected_equals

    def test__repr__(self):
        name, quantity = "Dolphin", 9
        egg = Egg(name, quantity=quantity)

        result: str = repr(egg)

        assert result == f"Egg({name}: {quantity})"

    @pytest.mark.parametrize("standard_egg_name", ["Wolf", "Fox", "FlyingPig", "Dragon"])
    def test_is_standard_egg(self, standard_egg_name: str):
        egg = Egg(standard_egg_name)

        assert egg.is_standard_egg() is True
        assert egg.is_quest_egg() is False

    @pytest.mark.parametrize("quest_egg_name", [
        "Dolphin", "Gryphon", "Deer", "Egg", "Seahorse", "Parrot"
    ])
    def test_is_quest_egg(self, quest_egg_name: str):
        egg = Egg(quest_egg_name)

        assert egg.is_standard_egg() is False
        assert egg.is_quest_egg() is True

    @pytest.mark.parametrize(
        "egg_name,potion_name",
        list(zip(EggData.drop_egg_names,
                 HatchPotionData.hatch_potion_names))
    )
    def test_can_be_hatched_by_standard_egg_true(self, egg_name: str,
                                                 potion_name: str):
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)

        # standard eggs can be hatched by any potion
        assert egg.can_be_hatched_by(potion) is True

    @pytest.mark.parametrize(
        "egg_name,potion_name",
        list(zip(EggData.quest_egg_names,
                 HatchPotionData.drop_hatch_potion_names))
    )
    def test_can_be_hatched_by_magic_egg_drop_potion_true(self,
                                                          egg_name: str,
                                                          potion_name: str):
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)

        # quest eggs can be hatched by standard potion
        assert egg.can_be_hatched_by(potion) is True

    @pytest.mark.parametrize(
        "egg_name,potion_name",
        list(zip(
            random.sample(EggData.quest_egg_names, k=_SAMPLE_SIZE),
            random.sample(HatchPotionData.non_drop_hatch_potion_names, k=_SAMPLE_SIZE)
        ))
    )
    def test_can_be_hatched_by_magic_egg_non_drop_potion_false(self,
                                                               egg_name: str,
                                                               potion_name: str):
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)

        # quest eggs cannot be hatched by non standard potion
        assert egg.can_be_hatched_by(potion) is False


class TestEggCollection:
    def test__init__empty__ok(self):
        collection = EggCollection()
        assert collection == EggCollection({})
        assert len(collection) == 0

    def test__init__ok(self):
        collection = EggCollection({"Wolf": 0, "Whale": 42, "Horse": 2})

        assert len(collection) == 3
        assert collection["Wolf"] == Egg("Wolf", quantity=0)
        assert collection["Whale"] == Egg("Whale", quantity=42)
        assert collection["Horse"] == Egg("Horse", quantity=2)

    def test__eq__(self):
        assert EggCollection() == EggCollection()
        assert EggCollection() == EggCollection({})
        assert EggCollection({"Cow": 1}) == EggCollection({"Cow": 1})
        assert EggCollection({"Cow": 1}) != EggCollection({"Cow": 2})
        assert EggCollection({"Cow": 1}) != EggCollection({"Rat": 1})
        assert EggCollection() != EggCollection({"Ferret": 1})

    def test_values_ok(self):
        collection = EggCollection({"Wolf": 1, "Whale": 41, "Horse": 2})

        generator = collection.values()
        assert next(generator) == Egg("Wolf", quantity=1)
        assert next(generator) == Egg("Whale", quantity=41)
        assert next(generator) == Egg("Horse", quantity=2)
        with pytest.raises(StopIteration):
            _ = next(generator)

    def test_values_as_list_ok(self):
        collection = EggCollection({"Wolf": 1, "Whale": 41, "Horse": 2})

        result: List[Egg] = list(collection.values())

        expected: List[Egg] = [
            Egg("Wolf", quantity=1), Egg("Whale", quantity=41), Egg("Horse", quantity=2)
        ]
        assert result == expected

    def test__iter__ok(self):
        collection = EggCollection({"Wolf": 1, "Whale": 42, "Horse": 2})

        iterator = iter(collection)

        assert next(iterator) == "Wolf"
        assert next(iterator) == "Whale"
        assert next(iterator) == "Horse"
        with pytest.raises(StopIteration):
            next(iterator)

    def test_remove_egg_not_available_faile(self):
        collection = EggCollection({"Whale": 1})

        not_found_egg = "Unicorn"
        with pytest.raises(EggException) as exec_info:
            collection.remove_egg(Egg(not_found_egg))

        expected_msg = f"{not_found_egg} was not in the collection "
        assert str(exec_info.value).startswith(expected_msg)

    def test_remove_egg_ok(self):
        egg1_name, egg1_quantity = "Wolf", 1
        egg2_name, egg2_quantity = "Whale", 42
        egg3_name, egg3_quantity = "Horse", 2
        collection = EggCollection(
            {egg1_name: egg1_quantity, egg2_name: egg2_quantity, egg3_name: egg3_quantity}
        )

        collection.remove_egg(Egg(egg1_name))
        collection.remove_egg(Egg(egg2_name))
        collection.remove_egg(Egg(egg3_name))

        assert collection[egg1_name] == Egg(egg1_name, quantity=egg1_quantity - 1)
        assert collection[egg2_name] == Egg(egg2_name, quantity=egg2_quantity - 1)
        assert collection[egg3_name] == Egg(egg3_name, quantity=egg3_quantity - 1)

    def test_get_standard_egg_collection_ok(self):
        egg1_name, egg1_quantity = "Wolf", 1
        egg2_name, egg2_quantity = "Whale", 42  # quest egg
        egg3_name, egg3_quantity = "Fox", 0
        egg4_name, egg4_quantity = "PandaCub", 69
        collection = EggCollection({
            egg1_name: egg1_quantity,
            egg2_name: egg2_quantity,
            egg3_name: egg3_quantity,
            egg4_name: egg4_quantity
        })

        result: EggCollection = collection.get_standard_egg_collection()

        expected = EggCollection({
            egg1_name: egg1_quantity, egg3_name: egg3_quantity, egg4_name: egg4_quantity
        })
        assert result == expected

    def test_get_quest_egg_collection_ok(self):
        egg1_name, egg1_quantity = "Wolf", 1  # not quest egg
        egg2_name, egg2_quantity = "Whale", 42
        egg3_name, egg3_quantity = "Horse", 0
        egg4_name, egg4_quantity = "Axolotl", 69
        collection = EggCollection({
            egg1_name: egg1_quantity,
            egg2_name: egg2_quantity,
            egg3_name: egg3_quantity,
            egg4_name: egg4_quantity
        })

        result: EggCollection = collection.get_quest_egg_collection()

        expected = EggCollection({
            egg2_name: egg2_quantity, egg3_name: egg3_quantity, egg4_name: egg4_quantity
        })
        assert result == expected



#!/usr/bin/env python3
import pytest
from click.testing import CliRunner, Result

from hopla.cli.get_user.stats import stats
from hopla.cli.groupcmds.get_user import HabiticaUser


class TestStatsCliCommand:
    MP = 213.1
    GP = 10.2
    LVL = 104

    def test_stats_all_ok(self, test_user: HabiticaUser):
        runner = CliRunner()

        result: Result = runner.invoke(stats, ["all"], obj=test_user)

        assert f'"mp": {TestStatsCliCommand.MP}' in result.stdout
        assert f'"gp": {TestStatsCliCommand.GP}' in result.stdout
        assert f'"lvl": {TestStatsCliCommand.LVL}' in result.stdout
        assert result.exit_code == 0

    def test_stats_all_same_as_no_arg_ok(self, test_user: HabiticaUser):
        runner = CliRunner()

        result_all: Result = runner.invoke(stats, ["all"], obj=test_user)
        result_no_arg: Result = runner.invoke(stats, obj=test_user)

        assert result_all.stdout == result_no_arg.stdout
        assert result_all.exit_code == result_no_arg.exit_code

    def test_stats_with_level_ok(self, test_user: HabiticaUser):
        runner = CliRunner()

        result_level: Result = runner.invoke(stats, ["level"], obj=test_user)

        assert result_level.stdout == f"{TestStatsCliCommand.LVL}\n"
        assert result_level.exit_code == 0

    @pytest.fixture
    def test_user(self) -> HabiticaUser:
        user_stats = {
            "mp": TestStatsCliCommand.MP,
            "gp": TestStatsCliCommand.GP,
            "lvl": TestStatsCliCommand.LVL
        }
        return HabiticaUser({
            "stats": user_stats,
            # contributor is just some extra data to make sure that we ignore this.
            "contributor": {"level": 3}}
        )



"""
The module with CLI code that handles the `hopla get-user inventory` command.
"""

import logging

import click

from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.cli.groupcmds.get_user import pass_user, HabiticaUser

log = logging.getLogger()

valid_item_groups = click.Choice([
    "pets", "mounts", "eggs", "food", "hatchingPotions",
    "gear", "quests", "currentPet", "currentMount",
    "lastDrop", "all"
])


def inventory_alias_to_official_habitica_name(inventory_name: str):
    """Take an inventory_name argument and return the canonical Habitica item name"""
    # pylint: disable=too-many-return-statements
    if inventory_name in ["hatchingpotions", "hatchingPotion"]:
        return "hatchingPotions"
    if inventory_name in ["pet"]:
        return "pets"
    if inventory_name in ["mount"]:
        return "mounts"
    if inventory_name in ["currentpet"]:
        return "currentPet"
    if inventory_name in ["currentmount"]:
        return "currentMount"
    if inventory_name in ["lastdrop"]:
        return "lastDrop"

    return inventory_name


@click.command(
    context_settings=dict(token_normalize_func=inventory_alias_to_official_habitica_name)
)
@click.argument("item_group_name", type=valid_item_groups, default="all")
@pass_user
def inventory(user: HabiticaUser, item_group_name: str) -> dict:
    """Get items from the user's inventory

    Get all items if no specific item group is specified.

    \f
    :param user:
    :param item_group_name: The type of items in the inventory (default: all)
    :return: The specified inventory
    """
    log.debug(f"hopla get-user inventory item_group={item_group_name}")

    data_items = user.get_inventory()
    if item_group_name == "all":
        requested_user_data = data_items
    else:
        requested_user_data = data_items[item_group_name]
    click.echo(JsonFormatter(requested_user_data).format_with_double_quotes())
    return requested_user_data



"""
The module with CLI code that handles the `hopla get-group command.
"""

import logging
from typing import Optional

import click
import requests

from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.outputformatter import JsonFormatter

log = logging.getLogger()


class HabiticaGroupRequest(HabiticaRequest):
    """An API request object for getting group information.

    Related URLs:
    [wiki on groups](https://habitica.fandom.com/wiki/Groups)
    [apidoc on groups](https://habitica.com/apidoc/#api-Group)
    """

    def __init__(self, group_id: Optional[str] = "party"):
        # "party" is the magic string that identifies your current party.
        self.group_id = group_id

    @property
    def path(self) -> str:
        """Get the path for the url"""
        return f"/groups/{self.group_id}"

    @property
    def url(self) -> str:
        """Get the path for the url"""
        return UrlBuilder(path_extension=self.path).url

    def get_group_request(self) -> requests.Response:
        """Perform a GET request to get a group information."""
        return requests.get(url=self.url, headers=self.default_headers)

    def get_group_data_or_exit(self) -> dict:
        """Get the group data or exit if the API request failed."""
        return get_data_or_exit(self.get_group_request())


@click.command()
@click.argument("group_id", default="party")
def get_group(group_id: str):
    """Get party information.

    GROUP_ID is the uuid of the group. When omitted your party information will be returned.

    \b
    Example
    ----
    # Get your party information:
    $ hopla get-group
    $ hopla get-group party

    \b
    # Get quest information
    $ hopla get-group | jq .quest

    \b
    # keep track of quest progress every 60 seconds
    $ watch --differences=permanent -n 60 'hopla get-group | jq .quest'

    \b
    # Get tavern information
    $ hopla get-group tavern

    \b
    # Specify a group by id
    $ hopla get-group 6fe49760-087f-4b08-9432-86fe77c2f1ef

    """
    log.debug(f"hopla get-party {group_id=}")

    group_request = HabiticaGroupRequest(group_id=group_id)
    group_info = group_request.get_group_data_or_exit()

    click.echo(JsonFormatter(group_info).format_with_double_quotes())
    return group_info



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla feed-all` command.
"""
import logging
import sys
from typing import Callable, List, NoReturn, Optional, Union

import click

from hopla.cli.groupcmds.get_user import HabiticaUser, HabiticaUserRequest
from hopla.hoplalib.throttling import ApiRequestThrottler
from hopla.hoplalib.zoo.foodmodels import FoodStockpile, FoodStockpileBuilder
from hopla.hoplalib.zoo.petcontroller import FeedPostRequester
from hopla.hoplalib.zoo.zoomodels import Zoo, ZooBuilder
from hopla.hoplalib.zoo.zoofeed_algorithms import ZooFeedAlgorithm, ZookeeperFeedPlan

log = logging.getLogger()


def __get_feed_plan_or_exit() -> Union[NoReturn, ZookeeperFeedPlan]:
    """Get the user and build the zookeeper plan"""
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    stockpile: FoodStockpile = FoodStockpileBuilder().user(user).build()
    zoo: Zoo = ZooBuilder(user).build()

    algorithm = ZooFeedAlgorithm(zoo=zoo, stockpile=stockpile)
    return algorithm.make_plan()


def __confirm_with_user_or_abort(plan: ZookeeperFeedPlan) -> Optional[NoReturn]:
    """Ask the user to confirm the specified plan.

    :param plan: the zookeeper feed plan to display
    :return: Don't return if the user wants to abort
    """

    prompt_msg = f"{plan.format_plan()}\nDo you want to proceed?"
    click.confirm(text=prompt_msg, abort=True)


def feed_all_pets_and_exit() -> NoReturn:
    """Feed all the pets"""
    plan: ZookeeperFeedPlan = __get_feed_plan_or_exit()
    if plan.isempty():
        click.echo(
            "The feed plan is empty. Reasons could include:\n"
            "1. There is insufficient food available to turn pets into mounts\n"
            "2. You don't have any feedable pets."
        )
        sys.exit(0)

    __confirm_with_user_or_abort(plan)

    feed_requests: List[Callable[[None], None]] = []
    for item in plan:
        feed_requester: FeedPostRequester = FeedPostRequester.build_from(item)
        feed_requests.append(feed_requester.post_feed_request_get_data_or_exit)

    throttler = ApiRequestThrottler(feed_requests)
    throttler.execute_all_requests()
    sys.exit(0)


@click.command()
def feed_all():
    """Feed all your pets.

    This command will first feed normal pets, then your quest pets, and
    finally all your pets that were hatched with magic hatching potions.

    Not this command will first show you the feed plan, and for safety,
    ask for your confirmation. Pets will only if you confirm this prompt.
    """
    log.debug("hopla feed-all")
    feed_all_pets_and_exit()



"""
Library code to handle Hopla's configuration.
"""
import logging
from configparser import ConfigParser
from pathlib import Path
from typing import List
from collections import namedtuple

from hopla.hoplalib.errors import PrintableException
from hopla.hoplalib.common import get_configuration_dirpath, EnvironmentVariables

log = logging.getLogger()


class HoplaConfigurationFile:
    """Hopla's configuration file"""

    def __init__(self, *, alternative_file=None):
        self.__alternative_file = alternative_file
        self.__global_env_var_hopla_conf_file = EnvironmentVariables.HOPLA_CONF_FILE

    @property
    def file_path(self) -> Path:
        """Get the hopla configuration file as a Path"""
        if self.__alternative_file is not None:
            config_file: Path = Path(self.__alternative_file)
        elif self.__global_env_var_hopla_conf_file is not None:
            config_file: Path = Path(self.__global_env_var_hopla_conf_file)
        else:
            config_file: Path = get_configuration_dirpath() / "hopla.conf"

        return config_file.resolve()

    def exists(self) -> bool:
        """Return true if the hopla configuration file exists"""
        return self.file_path.exists() and self.file_path.is_file()


class ConfigurationFileParser:
    """Class that writes and reads from the Hopla configuration file"""

    def __init__(self):
        self._conf_file = HoplaConfigurationFile()
        self.config_parser = ConfigParser()

    def get_full_config_name(self, full_config_name: str, *, fallback=None):
        """takes a full_config_name and returns the corresponding value in the config file

        For example: `get_full_config_name("cmd_all.loglevel")` returns 'warning' if
        the hopla config file contains:
        \b
        [cmd_all]
        loglevel = warning


        \f
        :param fallback:
        :param full_config_name:
        :return:
        """

        self.config_parser.read(self._conf_file.file_path)
        configuration_setting = FullConfigurationNameStr(full_config_name_str=full_config_name)

        return self.config_parser.get(
            section=configuration_setting.section,
            option=configuration_setting.short_config_name,
            fallback=fallback
        )

    def set_full_config_name(self, full_config_name: str, new_value):
        """
        Given a fully qualified config name (such cmd_all.loglevel) and a value, this
        function sets the new_value and returns the new_value.
        """
        self.config_parser.read(self._conf_file.file_path)
        configuration_setting = FullConfigurationNameStr(full_config_name_str=full_config_name)

        self.config_parser.set(
            section=configuration_setting.section,
            option=configuration_setting.short_config_name,
            value=new_value
        )
        with open(self._conf_file.file_path, mode="w", encoding="utf-8") as updated_conf_file:
            self.config_parser.write(updated_conf_file)

        return new_value


class InvalidFullNameFormat(PrintableException):
    """An exception that is thrown when a configuration full name doesn't have the valid format"""


ConfigurationName = namedtuple(typename="ConfigurationNameType",
                               field_names=["section", "config_name"])


class FullConfigurationNameStr:
    """The full configuration name of format:
            ini_section_name.key_name

        the full name 'cmd_all.loglevel' maps to the following in a config file:
        [cmd_all]
        loglevel = ...
    """

    def __init__(self, full_config_name_str: str):
        try:
            split_name: List[str] = full_config_name_str.split(".")
            self.__section, self.__config_short_name = split_name
            self.__config_name = ConfigurationName(section=self.__section,
                                                   config_name=self.__config_short_name)
        except ValueError as ex:
            raise InvalidFullNameFormat(
                "Expected a full_config_name of format: 'ini_section_name.key_name'"
                f"but received {full_config_name_str}"
            ) from ex

    def __str__(self) -> str:
        return str(self.get_validated_config_name())

    def get_validated_config_name(self) -> ConfigurationName:
        """Return the underlying named tuple"""
        return self.__config_name

    @property
    def section(self) -> str:
        """Return the section part of the Full configuration name string"""
        return self.__section

    @property
    def short_config_name(self):
        """Return the (short) name part of the Full configuration name string"""
        return self.__config_short_name


class HoplaDefaultConfiguration:
    """
    A class responsible for keeping track of the supported configuration and the default values.
    """

    @property
    def default_config_as_parser(self) -> ConfigParser:
        """
        Returns a ConfigParser with the default configuration assuming that nobody
        ever configured anything.

        :return:
        """
        default_config = ConfigParser()
        # [cmd_all] # config for all command
        # [cmd_XXX] # config for command XXX
        all_commands_section = "cmd_all"
        default_config.add_section(all_commands_section)

        # debug: for developers (too much info)
        # info: for developers (basic info, incl. graceful degradation)
        # warning: something worth of creating an issue on github, but nothing broke
        # error: user experienced something breaking down
        default_config.set(all_commands_section, "loglevel", "warning")
        return default_config

    def supported_sections(self):
        """Return the sections that are supported in the Hopla Configuration file."""
        return self.default_config_as_parser.sections()


class ConfigInitializer:
    """Helper class for initializing Hopla's configuration files."""

    def __init__(self):
        self.config_file = HoplaConfigurationFile()

    def initialize_before_running_cmds(self) -> bool:
        """"
        Create the default config file if no config file exists.

        :return: True if a file was created, false else.
        """
        if self.config_file.exists() is False:
            self._create_empty_config_file()
            default_config = HoplaDefaultConfiguration().default_config_as_parser
            with open(self.config_file.file_path, mode="w", encoding="utf-8") as new_conf_file:
                default_config.write(new_conf_file)
            return True

        return False

    def _create_empty_config_file(self):
        Path.mkdir(self.config_file.file_path.parent, parents=True, exist_ok=True)
        with open(self.config_file.file_path, mode="w", encoding="utf-8"):
            pass  # no need to write to it, just create it



#!/usr/bin/env python3
"""
A module with helper functions for the `hopla feed` and `hopla feed-all`
CLI commands.
"""
import sys
from typing import NoReturn, Union

import click
import requests

from hopla.hoplalib.outputformatter import JsonFormatter


def get_feed_data_or_exit(feed_response: requests.Response) -> Union[NoReturn, dict]:
    """
    Given a feed response, if the API request was successful, return interesting
    feed information. On failure, exit with an error message.
    """
    response_json = feed_response.json()
    if response_json["success"]:
        feed_data = {
            "feed_status": response_json["data"],
            "message": response_json["message"]
        }

        click.echo(JsonFormatter(feed_data).format_with_double_quotes())
        return feed_data

    click.echo(response_json["message"])
    sys.exit(1)



from unittest.mock import MagicMock, patch

import pytest
from _pytest.capture import CaptureResult
from click.testing import CliRunner, Result
from requests.status_codes import codes

from hopla.cli.feed import feed, get_appropriate_food_or_exit, get_feed_times_until_mount, \
    print_favorite_food_and_exit
from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.zoo.petmodels import Pet


class MockFeedResponse:
    def __init__(self, *, json, status_code: int = 200):
        self.__json = json
        self.status_code = status_code

    def json(self):
        return self.__json


class TestFeedCliCommand:
    # https://habitica.com/apidoc/#api-User-UserFeed

    @pytest.mark.parametrize(
        "pet_name,expected_favorite_food", [
            ("Wolf-Base", "Meat"),
            ("LionCub-SandSculpture", "Any"),
            ("Spider-White", "Milk"),
            ("Spider-Red", "Strawberry"),
            ("Spider-Shade", "Chocolate")
        ]
    )
    def test_list_favorite_food_ok(self, pet_name: str,
                                   expected_favorite_food: str):
        runner = CliRunner()
        result = runner.invoke(feed, [pet_name, "--list-favorite-food"])

        assert result.exit_code == 0
        assert result.output == f"{expected_favorite_food}\n"

    @patch("hopla.cli.feed.FeedPostRequester.post_feed_request")
    def test_feed_pet_once_ok(self, mock_feed_api_call: MagicMock):
        pet_name = "Wolf-Zombie"
        food_name = "RottenMeat"

        feed_status_expected = 15
        message_expected = f"{pet_name} really likes the Rotten Meat!"
        response_content = {"success": True, "data": feed_status_expected,
                            "message": message_expected, "userV": 59173,
                            "appVersion": "4.205.1"}

        mock_feed_api_call.return_value = MockFeedResponse(json=response_content)

        runner = CliRunner()
        result = runner.invoke(feed, [pet_name, food_name])

        mock_feed_api_call.assert_called_once_with()  # no args
        assert result.exit_code == 0
        assert f'"feed_status": {feed_status_expected},' in result.output
        assert f'"message": "{message_expected}"' in result.output

    @pytest.mark.parametrize(
        "pet_name,expected_favorite_food", [
            ("Dolphin-Red", "Strawberry"),
            ("Velociraptor-Shade", "Chocolate"),
            ("Hippo-Skeleton", "Fish"),
            ("Treeling-Zombie", "RottenMeat"),
            ("Pterodactyl-Desert", "Potatoe"),
        ]
    )
    @patch("hopla.cli.feed.FeedPostRequester.post_feed_request")
    def test_feed_pet_with_favorite_food_ok(self, mock_feed_api_call: MagicMock,
                                            pet_name: str,
                                            expected_favorite_food: str):
        feed_status_expected = 10
        # Side note: This message_expected variable is not perfect because the user' 'text'
        # naming for pets and foods differs often slightly from their 'key'
        # values (e.g. RottenMeat vs. Rotten Meat, DolphinRed vs. Red Dolphin). But that
        # is not relevant here, so we are happy anyways.
        message_expected = f'{pet_name} really likes the {expected_favorite_food}!'
        response_content = {'success': True, 'data': feed_status_expected,
                            'message': message_expected}
        mock_feed_api_call.return_value = MockFeedResponse(json=response_content)

        runner = CliRunner()
        result = runner.invoke(feed, [pet_name])  # We are not passing any food!

        mock_feed_api_call.assert_called_once_with()  # no args
        assert result.exit_code == 0
        assert f'"feed_status": {feed_status_expected},' in result.output
        assert f'"message": "{message_expected}"' in result.output

    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    @patch("hopla.cli.feed.FeedPostRequester.post_feed_request")
    def test_feed_pet_with_abundant_food(self, mock_feed_request: MagicMock,
                                         mock_user_request: MagicMock):
        most_abundant_food = "Meat"
        magic_pet_name = "PandaCub-Rainbow"

        mock_user_request.return_value = HabiticaUser({
            "items": {"food": {most_abundant_food: 10, "Honey": 4}}
        })

        expected_feed_data = 10
        expected_msg = f"{magic_pet_name} really likes the {most_abundant_food}!"
        success_response = {
            "success": True,
            "data": expected_feed_data,
            "message": expected_msg
        }
        mock_feed_request.return_value = MockFeedResponse(json=success_response)

        runner = CliRunner()
        result: Result = runner.invoke(feed, [magic_pet_name])

        assert result.exit_code == 0
        assert f'"feed_status": {expected_feed_data}' in result.stdout
        assert f'"message": "{expected_msg}"' in result.stdout

    @patch("hopla.cli.feed.FeedPostRequester.post_feed_request")
    def test_feed_pet_once_fail(self, mock_feed_api_call: MagicMock):
        pet_name = "LionCub-SandSculpture"
        food_name = "Strawberry"

        expected_errmsg = "You already have that mount. Try feeding another pet."

        response_content = {
            "success": False,
            "error": "NotAuthorized",
            "message": expected_errmsg
        }
        mock_feed_api_call.return_value = MockFeedResponse(json=response_content,
                                                           status_code=codes.unauthorized)
        runner = CliRunner()
        result: Result = runner.invoke(feed, [pet_name, food_name])

        mock_feed_api_call.assert_called_with()  # no args

        assert result.exit_code == 1
        assert result.stdout == f"{expected_errmsg}\n"

    def test_cannot_specify_times_and_until_mount(self):
        runner = CliRunner()
        result = runner.invoke(feed, ["Rat-Red", "--until-mount", "--times", 2])

        expected_errmsg = "Conflicting options: Cannot specify both --times and --until-mount"
        assert expected_errmsg in result.stdout
        assert result.exit_code == 2

    @patch("hopla.cli.feed.FeedPostRequester")
    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    def test_feed_until_mount(self,
                              mock_user_request: MagicMock,
                              mock_feed_requester: MagicMock):
        pet_name = "Egg-Shade"
        food_name = "Strawberry"
        feed_status = 47
        mock_user_request.return_value = HabiticaUser(
            {"items": {"pets": {pet_name: feed_status},
                       "mounts": {}}}
        )

        response_msg = f"You have tamed {pet_name}, let's go for a ride!"
        response_data = -1

        # This is not the prettiest ways of solving this. Feel free
        # to refactor this mocking logic.
        class MockRequester:
            class MockResponse:
                def json(self):
                    return {"success": True, "data": response_data,
                            "message": response_msg}

            def post_feed_request(self):
                return MockRequester.MockResponse()

        mock_feed_requester.return_value = MockRequester()

        runner = CliRunner()
        result: Result = runner.invoke(feed, [pet_name, food_name, "--until-mount"])

        expected_feed_times = 2
        mock_feed_requester.assert_called_with(
            pet_name=pet_name,
            food_name=food_name,
            food_amount=expected_feed_times
        )
        assert result.exit_code == 0
        assert response_msg in result.stdout
        assert f'"feed_status": {response_data}' in result.stdout


class TestPrintFavoriteFood:
    @pytest.mark.parametrize(
        "pet_name,expected_food", [
            ("Gryphon-Zombie", "RottenMeat"),
            ("Wolf-Vampire", "Any"),
            ("Phoenix-Base", "Unfeedable"),
            ("Gryphon-White", "Milk")
        ]
    )
    def test_print_favorite_food_and_exit(self, pet_name: str,
                                          expected_food: str,
                                          capsys):
        with pytest.raises(SystemExit):
            print_favorite_food_and_exit(pet_name)

        captured: CaptureResult = capsys.readouterr()
        assert captured.out == f"{expected_food}\n"
        assert captured.err == ""


class TestGetFeedTimesUntilMount:
    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    def test_get_feed_times_until_mount_ok(self, mock_user_request: MagicMock):
        pet_name = "Rat-Red"
        food_name = "Strawberry"
        feed_status = 21
        mock_user_request.return_value = HabiticaUser({"items": {"pets": {pet_name: feed_status},
                                                                 "mounts": {}}})

        result: int = get_feed_times_until_mount(pet_name, food_name)

        expected_strawberries = 6
        assert result == expected_strawberries

    PET_NAME = "Rat-White"

    @pytest.mark.parametrize("no_pet_user", [
        HabiticaUser({"items": {"pets": {}, "mounts": {}}}),
        HabiticaUser({"items": {"pets": {PET_NAME: 0}, "mounts": {}}}),
        HabiticaUser({"items": {"pets": {PET_NAME: 0}, "mounts": {PET_NAME: None}}}),
        HabiticaUser({"items": {"pets": {PET_NAME: -1}, "mounts": {}}}),
        HabiticaUser({"items": {"pets": {PET_NAME: -1}, "mounts": {PET_NAME: None}}}),
    ])
    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    def test_get_feed_times_until_mount_no_pet_fail(self,
                                                    mock_user_request: MagicMock,
                                                    no_pet_user: HabiticaUser):
        food_name = "Strawberry"
        mock_user_request.return_value = no_pet_user

        with pytest.raises(SystemExit) as execinfo:
            get_feed_times_until_mount(TestGetFeedTimesUntilMount.PET_NAME, food_name)

        expected_msg = (
            f"Can't feed pet {TestGetFeedTimesUntilMount.PET_NAME}. You don't have this pet."
        )
        assert str(execinfo.value) == expected_msg

    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    def test_get_feed_times_until_mount_have_mount_fail(self,
                                                        mock_user_request: MagicMock):
        name = "Rat-Red"
        food_name = "Strawberry"
        feed_status = 5
        mock_user_request.return_value = HabiticaUser(
            {"items": {"pets": {name: feed_status},
                       "mounts": {name: True}}}
        )

        with pytest.raises(SystemExit) as execinfo:
            get_feed_times_until_mount(name, food_name)

        err_msg = str(execinfo.value)
        assert err_msg == f"Can't feed pet {name}. You have the mount."

    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    def test_get_feed_times_until_mount_pet_unfeedable_data(self,
                                                            mock_user_request: MagicMock):
        pet_name = "Phoenix-Base"  # unfeedable
        food_name = "Strawberry"
        feed_status = 5
        mock_user_request.return_value = HabiticaUser(
            {"items": {"pets": {pet_name: feed_status},
                       "mounts": {}}}
        )

        with pytest.raises(SystemExit) as execinfo:
            get_feed_times_until_mount(pet_name, food_name)

        err_msg = str(execinfo.value)
        assert f"Can't feed pet {pet_name}." in err_msg


class TestGetAppropriateFoodOrExit:

    @patch("hopla.cli.feed.Pet")
    def test_get_appropriate_food_or_exit_bug_found(self, mock_pet: MagicMock):
        # This test tests the branch that should be unreachable if we configured
        # click correctly.
        pet_name = "ImpossiblePet-Base"

        class ImpossiblePet(Pet):
            def __init__(self, name):
                self.pet_name = name

            def has_just_1_favorite_food(self) -> bool:
                return False

            def likes_all_food(self) -> bool:
                return False

        mock_pet.return_value = ImpossiblePet(pet_name)

        with pytest.raises(YouFoundABugRewardError) as execinfo:
            get_appropriate_food_or_exit(pet_name)

        err_msg = str(execinfo.value)
        assert f"We tried to find the appropriate food for {pet_name=}.\n" in err_msg
        assert GlobalConstants.NEW_ISSUE_URL in err_msg



"""
Module with some Hopla common logic and data
"""
from typing import Final
from pathlib import Path
import os
import click


class GlobalConstants:
    """Class of global variables to be used by the entire application.

    Note that some of these constant may be upgraded to configuration parameters.
    """
    HABITICA_API_VERSION: Final[str] = "v3"
    """Version of the habitica API"""

    API_DOMAIN: Final[str] = "https://habitica.com"
    """Domain that provides the API """

    DEVELOPMENT_UUID: Final[str] = "79551d98-31e9-42b4-b7fa-9d89b0944319"
    """UUID of developer"""

    APPLICATION_NAME: Final[str] = "hopla"
    """Name of this application """

    X_CLIENT: Final[str] = f"{DEVELOPMENT_UUID}-{APPLICATION_NAME}"
    """
    BodyParam value used by hopla to identify itself to
    the habitica API <https://habitica.fandom.com/wiki/Guidance_for_Comrades>
    """

    SOURCE_CODE_REPOSITORY_URL: Final[str] = "https://github.com/melvio/hopla"
    """Repository for Hopla code"""

    ISSUE_URL: Final[str] = f"{SOURCE_CODE_REPOSITORY_URL}/issues"
    """Link which shows you an overview of the issues."""

    NEW_ISSUE_URL: Final[str] = f"{ISSUE_URL}/new"
    """Link which enables you to raise a new issue, suggest features, etc."""


class EnvironmentVariables:
    """Class with environment variables"""
    GLOBAL_ENV_VAR_HOPLA_AUTH_FILE: Final[str] = "HOPLA_AUTH_FILE"
    """ environment variable that a user can set to overwrite the
        default authenticate file """

    HOPLA_AUTH_FILE = os.environ.get(GLOBAL_ENV_VAR_HOPLA_AUTH_FILE)

    GLOBAL_ENV_VAR_HOPLA_CONF_FILE = "HOPLA_CONF_FILE"
    """ environment variable that a user can set to overwrite the
        default config file """

    HOPLA_CONF_FILE = os.environ.get(GLOBAL_ENV_VAR_HOPLA_CONF_FILE)


def get_configuration_dirpath() -> Path:
    """
    Get the most appropriate location for configuration (this is different per OS/environment)
    """
    return Path(click.get_app_dir(GlobalConstants.APPLICATION_NAME)).resolve()



#!/usr/bin/env python3
"""
Module with egg and hatch potion data.
"""
from typing import List


class EggData:
    """Class with data about eggs."""

    drop_egg_names = [
        "BearCub", "Cactus", "Dragon", "FlyingPig", "Fox", "LionCub",
        "PandaCub", "TigerCub", "Wolf"
    ]
    """
    Drop eggs. This is list was retrieved using: `hopla api content | jq .dropEggs | jq. keys`
    """

    quest_egg_names = [
        "Alligator", "Armadillo", "Axolotl", "Badger", "Beetle", "Bunny",
        "Butterfly", "Cheetah", "Cow", "Cuttlefish", "Deer", "Dolphin", "Egg",
        "Falcon", "Ferret", "Frog", "Gryphon", "GuineaPig", "Hedgehog",
        "Hippo", "Horse", "Kangaroo", "Monkey", "Nudibranch", "Octopus",
        "Owl", "Parrot", "Peacock", "Penguin", "Pterodactyl", "Rat", "Robot",
        "Rock", "Rooster", "Sabretooth", "SeaSerpent", "Seahorse", "Sheep",
        "Slime", "Sloth", "Snail", "Snake", "Spider", "Squirrel", "TRex",
        "Treeling", "Triceratops", "Turtle", "Unicorn", "Velociraptor",
        "Whale", "Yarn"
    ]
    """
    Quest Eggs. This is list was retrieved using: `hopla api content | jq .questEggs | jq keys`
    """

    egg_names: List[str] = drop_egg_names + quest_egg_names
    """All eggs in habitica."""


class HatchPotionData:
    """Class with data about hatching potions."""
    drop_hatch_potion_names = [
        "Base", "CottonCandyBlue", "CottonCandyPink", "Desert", "Golden",
        "Red", "Shade", "Skeleton", "White", "Zombie"
    ]
    """
    Drop hatching potions. This list was retrieved by using:
    hopla api content | jq.dropHatchingPotions | jq keys
    """

    magic_hatch_potion_names = [
        "Amber", "Aquatic", "Aurora", "AutumnLeaf", "BirchBark", "BlackPearl",
        "Bronze", "Celestial", "Cupid", "Ember", "Fairy", "Floral",
        "Fluorite", "Frost", "Ghost", "Glass", "Glow", "Holly", "IcySnow",
        "Moonglow", "MossyStone", "Peppermint", "PolkaDot", "Rainbow",
        "RoseQuartz", "RoyalPurple", "Ruby", "SandSculpture", "Shadow",
        "Shimmer", "Silver", "SolarSystem", "Spooky", "StainedGlass",
        "StarryNight", "Sunset", "Sunshine", "Thunderstorm", "Turquoise",
        "Vampire", "Watery", "Windup"
    ]
    """
    Magic hatching potions. This list was retrieved by using:
    hopla api content | jq .premiumHatchingPotions | jq keys
    """

    wacky_hatch_potion_names = ["Dessert", "Veggie"]
    """Wacky hatching potions. This list was retrieved by using:
           hopla api content | jq '.wackyHatchingPotions|keys'
    """
    non_drop_hatch_potion_names: List[str] = (
            magic_hatch_potion_names
            + wacky_hatch_potion_names
    )

    hatch_potion_names: List[str] = drop_hatch_potion_names + non_drop_hatch_potion_names
    """All potion names in habitica."""



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla hatch standard-egg` command.
"""
import logging
from typing import NoReturn

import click

from hopla.cli.groupcmds.hatch import hatch_egg
from hopla.hoplalib.hatchery.hatchdata import EggData, HatchPotionData

log = logging.getLogger()


@click.command()
@click.argument(
    "egg_name", metavar="STANDARD_EGG_NAME",
    type=click.Choice(EggData.drop_egg_names)
)
@click.argument(
    "potion_name", metavar="POTION_NAME",
    type=click.Choice(sorted(HatchPotionData.hatch_potion_names))
)
def standard_egg(egg_name: str, potion_name: str) -> NoReturn:
    """Hatch a standard egg.

    \b
    STANDARD_EGG_NAME =
        Name of the standard egg. Must be one of: BearCub, Cactus, Dragon,
        FlyingPig, Fox, LionCub, PandaCub, TigerCub, Wolf
    \b
    POTION_NAME =
        Name of the hatching potion. (e.g. White, Desert, Sunset, Windup)

    \b
    Examples
    -----
    # hatch a Wolf-Shade
    $ hopla hatch standard-egg Wolf Shade

    \b
    # hatch a Fox-SolarSystem
    $ hopla hatch standard-egg Fox SolarSystem

    \f
    [APIDOCS](https://habitica.com/apidoc/#api-User-UserHatch)
    :param egg_name:
    :param potion_name:
    :return:
    """
    log.debug(f"hopla hatch standard-egg {egg_name=} {potion_name=}")

    hatch_egg(egg_name=egg_name, potion_name=potion_name)



from pathlib import Path
from hopla.cli.complete import Bash


def test_default_bash_config_file_is_bashrc():
    bash = Bash()
    assert bash.config_file == Path.home() / ".bashrc"


def test_bash_enablement_code_for_bashrc():
    bash = Bash()
    assert bash.get_enablement_code_for_bashrc() == \
           'eval "$(_HOPLA_COMPLETE=bash_source hopla)"'


def test_bash_has_autocomplete_cmd():
    bash = Bash()
    assert bash.get_generated_autocomplete_cmd() == \
           "_HOPLA_COMPLETE=bash_source hopla"


def test_bas_supports_enablement():
    bash = Bash()
    assert bash.supports_automatic_complete_enablement()



#!/usr/bin/env python3
"""
Module with hopla versioning logic.
"""
from dataclasses import dataclass
from typing import Final, Optional

MAJOR_VERSION: Final[int] = 0
MINOR_VERSION: Final[int] = 0
PATCH_VERSION: Final[int] = 21
PRE_RELEASE: Final[Optional[str]] = "alpha"


@dataclass
class HoplaVersion:
    """Class with HoplaVersion constants."""

    def __init__(self,
                 major_version: int = MAJOR_VERSION,
                 minor_version: int = MINOR_VERSION,
                 patch_version: int = PATCH_VERSION,
                 pre_release: Optional[str] = PRE_RELEASE):
        self.major_version = major_version
        self.minor_version = minor_version
        self.patch_version = patch_version
        self.pre_release: str = pre_release or ""

    def semantic_version(self) -> str:
        """
        Returns the hopla version (Not to be confused with the habitica API version)

        The output satisfies [semantic versioning](https://semver.org/)

        Partial (Top-Level) BNF for valid SemVer versions:
            <valid semver> ::= <version core>
                             | <version core> "-" <pre-release>
                             | <version core> "+" <build>
                             | <version core> "-" <pre-release> "+" <build>

        :return: hopla version string
        """

        version_core = f"{self.major_version}.{self.minor_version}.{self.patch_version}"
        sep = "-" if self.pre_release != "" else ""

        return f"{version_core}{sep}{self.pre_release}"



#!/usr/bin/env python3
"""
The module with controllers for habitica tasks.
"""
import requests

from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.tasks.taskmodel import HabiticaTodo


class AddTodoRequest(HabiticaRequest):
    """An object that can perform an add To-Do request"""

    def __init__(self, habitica_todo: HabiticaTodo):
        self.url: str = UrlBuilder(path_extension="/tasks/user").url
        self.habitica_todo = habitica_todo

    def post_add_todo_request(self):
        """Perform the add To-Do request and return the data in case of success"""
        response = requests.post(
            url=self.url,
            headers=self.default_headers,
            json=self.habitica_todo.to_json_dict()
        )
        return get_data_or_exit(response)



"""
A helper module for Pet logic.
"""
from dataclasses import dataclass
from typing import Optional

from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.errors import PrintableException, YouFoundABugRewardError
from hopla.hoplalib.hatchery.hatchdata import HatchPotionData
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petdata import PetData


class InvalidPet(PrintableException):
    """Exception raised when a pet is invalid."""

    def __init__(self, msg: str, *, pet: Optional["Pet"] = None):
        super().__init__(msg)
        self.pet = pet


@dataclass
class Pet:
    """A habitica pet"""

    # pylint: disable=too-many-public-methods
    #################################################################
    # In a way, pylint is right. This class does have a lot of functions.
    # Most of these are oneliners, however we can also use polymorphism.
    # That would make it easier to find functions in this class.
    # We can use a PetFactory that creates QuestPets, MagicPets etc.
    # However, that is really overkill for now and not urgent enough.
    # Therefore, too-many-public-methods is disabled for this class.
    #################################################################

    def __init__(self, pet_name: str, *,
                 feed_status: FeedStatus = FeedStatus(5)):
        if pet_name not in PetData.pet_names:
            raise InvalidPet(f"{pet_name=} is not recognized by hopla.\n"
                             "Potential causes: \n"
                             "* did you spell it correctly?\n"
                             "* Is your pet relatively new? If so, please raise\n"
                             f" an issue at {GlobalConstants.NEW_ISSUE_URL}")

        self.name = pet_name
        self.feed_status = feed_status

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self.feed_status})"

    @property
    def hatch_potion_name(self) -> Optional[str]:
        """The hatching potion used to hatch the egg this pet came from."""
        if self.name in PetData.rare_pet_names:
            return None

        # This logic might break, but seems to be solid for past few years due to
        # stabling naming convention by the Habitica API developers.
        _, potion = self.name.split("-")
        return potion

    def is_available(self) -> bool:
        """Return True if the feed status says that the pet is available."""
        return self.feed_status.is_pet_available()

    def is_feedable(self) -> bool:
        """Return True if a pet cannot be fed at all."""
        return self.name not in PetData.unfeedable_pet_names

    def has_just_1_favorite_food(self) -> bool:
        """Return True if this pet likes only 1 type of food."""
        return self.name in PetData.only_1favorite_food_pet_names

    def likes_all_food(self) -> bool:
        """Return True if this pet prefers all food."""
        return self.name in PetData.magic_potion_pet_names

    def feed_status_explanation(self) -> str:
        """Explain the feed status of a pet."""
        if self.is_feedable() is False:
            return f"{self.name} is an unfeedable pet."
        if int(self.feed_status) == 0:
            return f"You released {self.name}, so you cannot feed it."

        if int(self.feed_status) == -1:
            return f"You can't feed {self.name=} you only have the mount"

        if int(self.feed_status) == 5:
            # remark: if we really want to know more, we can query the
            #         API to check if we have the self.name in the
            #         Users: jq .data.items.mounts
            return (f"Cannot determine if {self.name=} can be fed. \n"
                    "You either have:\n"
                    "1. Both the pet and mount. In this case, you cannot feed the pet \n"
                    "2. A pet that hasn't been fed but you don't have \n"
                    "   the mount. In this case, you can feed your pet")

        if int(self.feed_status) < 50:
            return f"{self.name=} can be fed"

        raise InvalidPet(f"Did not expect {self.feed_status=}. \n"
                         f"Looks like we failed to validate the Pet properly.",
                         pet=self)  # pragma: no cover

    def favorite_food(self, *,
                      default_value_for_unfeedable: str = "Unfeedable",
                      default_value_for_all_favorite_food: str = "Any") -> str:
        """Return the favorite food of this pet."""
        if self.is_feedable() is False:
            return default_value_for_unfeedable

        if self.name in PetData.magic_potion_pet_names:
            return default_value_for_all_favorite_food

        if self.has_just_1_favorite_food():
            return (FoodData.hatch_potion_favorite_food_mapping
                    .get(self.hatch_potion_name))

        raise InvalidPet(f"Could not find the feed habits of this {self.name=}",
                         pet=self)  # pragma: no cover

    def is_favorite_food(self, food_name: str) -> bool:
        """Return true if 'food_name' is this Pets favorite food."""
        if self.is_feedable() is False:
            return False
        if self.likes_all_food():
            return True
        if self.has_just_1_favorite_food():
            return self.favorite_food() == food_name

        raise InvalidPet(f"Unable to determine if {food_name=} is {self.name=}'s favorite.",
                         pet=self)  # pragma: no cover

    def required_food_items_until_mount(self, food_name: str) -> int:
        """
        Return the number of times food_name needs to be fed
        for this pet to turn into a mount
        """
        is_favorite: bool = self.is_favorite_food(food_name)
        return self.feed_status.required_food_items_to_become_mount(is_favorite)

    def is_generation1_pet(self) -> bool:
        """Return True if this pet is from the generation 1 pet"""
        return self.name in PetData.generation1_pet_names

    def is_quest_pet(self) -> bool:
        """
        Return True if this pet is a quest pet. This doesn't include
        special pets such as world event related pets.
        """
        return self.name in PetData.quest_pet_names

    def is_magic_hatch_pet(self) -> bool:
        """Return True if this pet is hatched from a magic potion."""
        return self.name in PetData.magic_potion_pet_names

    def is_from_drop_hatch_potions(self) -> bool:
        """
        Return True if the pet was hatched from one of the 'ordinary'
        potions. (Such as: Base, Desert, ...).
        """
        return self.hatch_potion_name in HatchPotionData.drop_hatch_potion_names


@dataclass
class Mount:
    """ The model class for mounts."""

    def __init__(self, mount_name: str, *,
                 availability_status: Optional[bool]):
        """
        Create a mount.

        Note that, in contrast to Pet.__init__, currently, there is no name
        correctness checking. Related ticket:
        https://github.com/melvio/hopla/issues/164

        :param mount_name:
        :param availability_status:
        """
        self.name = mount_name
        self._availability_status = availability_status

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self._availability_status})"

    def is_available(self) -> bool:
        """Return true if the user has the mount right now."""
        return self._availability_status is True


class InvalidPetMountPair(YouFoundABugRewardError):
    """Exception raised when an invalid PetMountPair is created."""


@dataclass
class PetMountPair:
    """A pair of a pet and its mount."""

    def __init__(self, *, pet: Optional[Pet],
                 mount: Optional[Mount]):
        """
        Create a PetMountPair.
        Raise an exception if the names of the pet and mount do not match.
        """
        if pet and mount:
            if pet.name != mount.name:
                msg = f"pet name '{pet.name}' must equal mount name '{mount.name}'"
                raise InvalidPetMountPair(msg)

        self.pet = pet
        self.mount = mount

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pet={self.pet}, mount={self.mount})"

    def can_feed_pet(self) -> bool:
        """Return true if the pet itself is feedable and there is no mount yet."""
        return (
                self.pet_available()
                and self.pet.is_feedable()
                and not self.mount_available()
        )

    def mount_available(self) -> bool:
        """True if the mount is in the pair"""
        return self.mount is not None and self.mount.is_available()

    def pet_available(self) -> bool:
        """True if the pet is in the pair"""
        return self.pet is not None and self.pet.is_available()



"""
Module with data about Habitica food information.
"""


class FoodData:
    """
    Helper class with valid Habitica food and drop hatching potions and
    the relationship between them.
    """

    hatch_potion_favorite_food_mapping = {
        "Base": "Meat",
        "White": "Milk",
        "Desert": "Potatoe",
        "Red": "Strawberry",
        "Shade": "Chocolate",
        "Skeleton": "Fish",
        "Zombie": "RottenMeat",
        "CottonCandyPink": "CottonCandyPink",
        "CottonCandyBlue": "CottonCandyBlue",
        "Golden": "Honey"
    }
    """
    Note: Pets hatched with magic hatching potions prefer any type of food.
    These pets are not supported by this dict.
    Rare favorite foods are also not supported such as Cake, Candy, and Pie.

    @see:
    * <https://habitica.fandom.com/wiki/Food_Preferences>
    * hopla api content | jq .dropHatchingPotions
    * hopla api content | jq .dropEggs

    also interesting:
    * hopla api content | jq .questEggs
    * hopla api content | jq .hatchingPotions
    * hopla api content | jq .premiumHatchingPotions
    * hopla api content | jq. wackyHatchingPotions
    """

    drop_food_names = list(hatch_potion_favorite_food_mapping.values())
    """A list of food items that can be dropped by doing tasks.

    These dont include cakes etc., those are rare collectibles.

    # for more info @see:
    #    hopla api content | jq .food
    #    hopla api content | jq '.food[] | select(.canDrop==true)'
    #    hopla api content | jq '[.food[] | select(.canDrop==true) | .key]'
    """



"""
Module that sets __version__ and then calls the kickstarter.
"""

from hopla.hoplalib.hoplaversion import HoplaVersion

try:
    # F401=module-imported-but-unused: hopla.kickstart.hopla is indeed
    # imported but not use directly. However, we still need it for
    # need it for build the right now.
    from hopla.kickstart import hopla  # noqa: F401
    from hopla.kickstart import kickstart_hopla

except ModuleNotFoundError:
    pass
    # This error catching is just here for the `python -m build` call that
    # fails to find click at build time when trying to find the
    # __version__ variable.

__version__ = HoplaVersion().semantic_version()
""" This __version__ is read by setup.cfg. """


def setup_hopla_application():
    """
    This function only calls kickstart_hopla. This reduces the
    required logic in base __init__.py
    """
    kickstart_hopla()



"""
Module with commonly used python-click functionality.
"""
import logging
import datetime
from typing import Optional, Any

import click

log = logging.getLogger()


class EnhancedDate(click.DateTime):
    """EnhancedDate adds 'today' and 'tomorrow' keywords to the click.DateTime type."""
    name = "enhanceddate"

    def __init__(self):
        super().__init__(formats=["%Y-%m-%d", "%d-%m-%Y"])

        # This will add 'today' and 'tomorrow to the help message
        self.formats += ["today", "tomorrow"]

    def convert(self, value: Any,
                param: Optional[click.Parameter],
                ctx: Optional[click.Context]) -> Any:
        """Convert a value provided by the user into a datetime object."""
        if value == "today":
            return datetime.date.today()
        if value == "tomorrow":
            return datetime.date.today() + datetime.timedelta(days=1)

        return super().convert(value, param, ctx)



"""
The module with CLI code that handles the `hopla get-user user-authenticate` command.
"""

import logging

import click

from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.cli.groupcmds.get_user import pass_user, HabiticaUser

log = logging.getLogger()

__SIGN_IN_OPTIONS = ["local", "google", "apple", "facebook"]
valid_auth_info_names = click.Choice([
    "username", "email", "profilename", *__SIGN_IN_OPTIONS, "all"
])


def auth_alias_to_official_habitica_name(auth_info_name: str):
    """
    Function that returns habitica official names for the CLI cmd:
    hopla get_user user-authenticate [alias]
    """
    if auth_info_name in ["e-mail", "mail"]:
        return "email"
    return auth_info_name


@click.command(
    context_settings=dict(token_normalize_func=auth_alias_to_official_habitica_name)
)
@click.argument("auth_info_name", type=valid_auth_info_names, default="all")
@pass_user
def auth(user: HabiticaUser, auth_info_name: str):
    """Get user authentication and identification info

    \b
    Examples
    ---
    \b
    # get a user's "local" (email based) sign in information
    $ hopla get-user auth local

    \b
    # get a user's apple/facebook/google sign in information
    $ hopla get-user auth apple
    $ hopla get-user auth facebook
    $ hopla get-user auth google

    \b
    # get the user email (assumes email based login)
    $ hopla get-user auth email


    """
    log.debug(f"hopla get-user auth auth_name={auth_info_name}")
    auth_data: dict = user.get_auth()

    if auth_info_name == "profilename":
        requested_data = user["profile"]["name"]
    elif auth_info_name == "username":
        requested_data = auth_data["local"]["username"]
    elif auth_info_name == "email":
        requested_data = auth_data["local"]["email"]
    elif auth_info_name in __SIGN_IN_OPTIONS:
        requested_data = auth_data[auth_info_name]
    else:
        requested_data = auth_data

    click.echo(JsonFormatter(requested_data).format_with_double_quotes())
    return requested_data



"""
Library code to help with Habitica API HTTPS-requests
"""
from dataclasses import dataclass

from hopla.hoplalib.authorization import AuthorizationHandler
from hopla.hoplalib.common import GlobalConstants


class RequestHeaders:
    """
    Helper class that takes care of HTTP request headers when interacting with
    the habitica API.

    For more information, see <https://habitica.fandom.com/wiki/Guidance_for_Comrades>
    """
    CONTENT_TYPE_HEADER_NAME = "Content-Type"
    CONTENT_TYPE_HEADER_VALUE_APPLICATION_JSON = "application/json"
    X_CLIENT_HEADER_NAME = "x-client"
    X_API_USER_HEADER_NAME = "x-api-user"
    X_API_KEY_HEADER_NAME = "x-api-key"

    def __init__(self, auth_parser: AuthorizationHandler = None):
        if auth_parser:
            self.hopla_auth_parser = auth_parser
        else:
            self.hopla_auth_parser = AuthorizationHandler()

    def get_default_request_headers(self) -> dict:
        """Return a dict of request headers that are used for nearly every
        habitica API request"""
        return {
            RequestHeaders.CONTENT_TYPE_HEADER_NAME:
                RequestHeaders.CONTENT_TYPE_HEADER_VALUE_APPLICATION_JSON,
            RequestHeaders.X_CLIENT_HEADER_NAME: GlobalConstants.X_CLIENT,
            RequestHeaders.X_API_USER_HEADER_NAME: self.hopla_auth_parser.user_id,
            RequestHeaders.X_API_KEY_HEADER_NAME: self.hopla_auth_parser.api_token
        }


class UrlBuilder:
    """
    Helper class for building habitica API URLs.
    """

    def __init__(self, *,
                 domain: str = GlobalConstants.API_DOMAIN,
                 api_version: str = GlobalConstants.HABITICA_API_VERSION,
                 path_extension: str = ""):
        self.domain = domain
        self.api_version = api_version
        self.path_extension = path_extension

    def __str__(self) -> str:
        return f"UrlBuilder(url={self.url})"

    def _get_base_url(self) -> str:
        return f"{self.domain}/api/{self.api_version}"

    @property
    def url(self) -> str:
        """Get the build URL"""
        return f"{self._get_base_url()}{self.path_extension}"


@dataclass(repr=True, init=False)
class HabiticaRequest:
    """
    A generic API request class with inheritable logic for specific
    request classes.
    """

    @property
    def default_headers(self):
        """
        Return the default headers with the user's credentials and the x-client header.
        """
        return RequestHeaders().get_default_request_headers()



"""
Library code that helps with outputting strings to the CLI user.
"""
from dataclasses import dataclass
import json


@dataclass
class JsonFormatter:
    """
    Class for formatting JSON s.t. it can be used in bash pipelines.
    """

    def __init__(self, json_as_dict: dict):
        self.json_as_dict = json_as_dict

    def format_with_double_quotes(self, indent: int = 2) -> str:
        """format a dictionary such that it has double quotes
        Double quoting allow the string to be used in pipelines

        :param indent:
        :return:
        """
        return json.dumps(self.json_as_dict, indent=indent)



#!/usr/bin/env python3
"""
Module that talks to the Habitica API to hatch eggs.
"""
from dataclasses import dataclass
from typing import Any, Dict

import requests
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder


@dataclass
class HatchRequester(HabiticaRequest):
    """HatchRequester that POSTs to hatch endpoint.

    [APIDOCS](https://habitica.com/apidoc/#api-User-UserHatch)
    """
    egg_name: str
    hatch_potion_name: str

    @property
    def url(self) -> str:
        """Hatch endpoint with egg and hatch_potion_name as path params."""
        path_extension = f"/user/hatch/{self.egg_name}/{self.hatch_potion_name}"
        return UrlBuilder(path_extension=path_extension).url

    def post_hatch_egg_request(self) -> requests.Response:
        """Perform a POST request on the Habitica API to hatch an egg."""
        return requests.post(url=self.url, headers=self.default_headers)

    def post_hatch_egg(self) -> str:
        """Perform a POST request on the Habitica API to hatch an egg, and
        return a str that can be nicely printed to the terminal.
        """
        response: requests.Response = self.post_hatch_egg_request()
        response_json: Dict[str, Any] = response.json()
        if response_json["success"] is True:
            return f"Successfully hatched a {self.egg_name}-{self.hatch_potion_name}."
        return f"{response_json['error']}: {response_json['message']}"



"""
A helper module for feeding logic.
"""
import copy
import math
from dataclasses import dataclass
from typing import Dict, Final, Optional

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.errors import PrintableException
from hopla.hoplalib.zoo.fooddata import FoodData


class InvalidFeedStatus(PrintableException):
    """Exception raised when a pet is invalid."""

    def __init__(self, msg: str, *, pet=None):
        super().__init__(msg)
        self.pet = pet


class FeedStatus:
    """A class implementing feed status logic for pets.

    Every freshly hatched pet starts at 5.
    50 would turn the pet into a mount, so 50 is impossible to reach.
    The feed status of 0 means that you had the pet before, but
    you released it.
    """
    PET_GREW_UP_TO_MOUNT: Final[int] = -1
    PET_RELEASED: Final[int] = 0

    START_FEED_STATE: Final[int] = 5
    MAX_FED_STATE: Final[int] = 49
    FULLY_FED_STATE: Final[int] = 50

    FAVORITE_INCREMENT: Final[int] = 5
    NON_FAVORITE_INCREMENT: Final[int] = 2

    def __init__(self, feed_status: int = START_FEED_STATE):
        invalid_status = (feed_status < FeedStatus.PET_GREW_UP_TO_MOUNT
                          or feed_status in [1, 2, 3, 4]
                          or feed_status > FeedStatus.MAX_FED_STATE)
        if invalid_status:
            raise InvalidFeedStatus(f"{feed_status=} is invalid")

        self.__feed_status = feed_status

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__feed_status})"

    def __eq__(self, other):
        return isinstance(other, FeedStatus) and int(other) == int(self)

    def __hash__(self):
        return hash(self.__feed_status)

    def __int__(self) -> int:
        return self.__feed_status

    def is_pet_available(self):
        """Return True if the user currently has this pet."""
        return self.__feed_status not in [
            FeedStatus.PET_GREW_UP_TO_MOUNT, FeedStatus.PET_RELEASED
        ]

    def required_food_items_to_become_mount(self, is_favorite_food: bool) -> int:
        """Return how many items of food we need to give to turn a pet into a mount."""
        target = FeedStatus.FULLY_FED_STATE - self.__feed_status
        if is_favorite_food:
            required_food = math.ceil(target / FeedStatus.FAVORITE_INCREMENT)
        else:
            required_food = math.ceil(target / FeedStatus.NON_FAVORITE_INCREMENT)
        return required_food

    def to_percentage(self) -> int:
        """
        Turn feed status into percentage understandable by the
        website user.
        <https://habitica.fandom.com/wiki/Food_Preferences>
        """
        if self.__feed_status == -1:
            return 100  # The pet is now a mount
        return self.__feed_status * 2


class FoodException(PrintableException):
    """Exception thrown when there is a food error."""

    def __init__(self, msg: str, *, food: Optional["Food"] = None):
        super().__init__(msg)
        self.food = food


@dataclass(frozen=True)
class Food:
    """A stockpile food item."""
    name: str

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.name})"

    def is_rare_food_item(self) -> bool:
        """
        Return False if the food_item item is a drop food.
        Otherwise, return True.
        """
        return self.name not in FoodData.drop_food_names


@dataclass
class FoodStockpile:
    """The food of a user."""

    def __init__(self, __stockpile: Dict[str, int]):
        self.__stockpile = __stockpile

    def __eq__(self, other):
        # pylint: disable=protected-access
        return isinstance(other, FoodStockpile) and self.__stockpile == other.__stockpile

    def __hash__(self):
        return hash(self.__stockpile)

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.__stockpile})"

    def add_food(self, food_name: str, *, n: int) -> "FoodStockpile":
        """
        Change the number of specified food in the stockpile.

        :param food_name: the food to modify
        :param n: When n is positive, add this number of food.
                  When negative, subtract n from this food from the stockpile.
        :return: The modified FoodStockPile
        """
        food = Food(food_name)
        if food.is_rare_food_item():
            raise FoodException(msg=f"Not Supported: {food_name=} is not supported.",
                                food=food)

        cur_quantity = self.__stockpile[food_name]

        if cur_quantity + n < 0:
            msg = (f"Insufficient food: Cannot remove {n=} of food from the stockpile\n"
                   f"The current quantity of {food_name} is {cur_quantity}.")
            raise FoodException(msg, food=Food(food_name))

        self.__stockpile[food_name] += n

        return self

    def add_food_dict(self, food_dict: Dict[str, int]) -> "FoodStockpile":
        """
        Add an entire food dict to the stockpile.

        Note that negative values will reduce the number of food items in this
        FoodStockPile

        :param food_dict:
        :return: the modified FoodStockPile
        """
        for food_name, times in food_dict.items():
            self.add_food(food_name, n=times)

        return self

    def get_most_abundant_food(self) -> str:
        """
        Return the most abundant food in the stockpile. If multiple food items
        occur equally most frequent, return one of them. This function makes
        no guarantee which is returned in that case.
        """
        return max(self.__stockpile, key=self.__stockpile.get)

    def has_sufficient(self, food_name: str, *, n: int) -> bool:
        """Return True if the stockpile has >=n of the specified food item."""
        return self.__stockpile[food_name] >= n

    def has_sufficient_abundant_food(self, *, n: int):
        """Return True if the stockpile has >=n of the most abundant food item."""
        food_name: str = self.get_most_abundant_food()
        return self.has_sufficient(food_name, n=n)

    def as_dict(self) -> Dict[str, int]:
        """Return a deep copy of the underlying data."""
        return copy.deepcopy(self.__stockpile)


class FoodStockpileBuilder:
    """
    Class that creates a FoodStockpile using the builder
    design pattern.
    """

    def __init__(self):
        self.__stockpile: Dict[str, int] = self.__empty_stockpile()

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.__stockpile})"

    def user(self, habitica_user: HabiticaUser) -> "FoodStockpileBuilder":
        """
        Add a user's food to the stockpile. This ignores saddles,
        cakes, and other rare foods.
        :param habitica_user:
        :return: self
        """
        food: Dict[str, int] = habitica_user.get_food()
        for food_name, quantity in food.items():
            if Food(food_name).is_rare_food_item() is False:
                self.__stockpile[food_name] = quantity
        return self

    def build(self) -> FoodStockpile:
        """
        Build a normalized FoodStockpile. This stockpile never has rare items, and
        it add a 0 count for foods that we don't have.
        """
        return FoodStockpile(self.__stockpile)

    @staticmethod
    def __empty_stockpile() -> Dict[str, int]:
        no_food = 0
        return dict.fromkeys(FoodData.drop_food_names, no_food)

    @classmethod
    def empty_stockpile(cls) -> FoodStockpile:
        """Create an empty stockpile."""
        return FoodStockpile(FoodStockpileBuilder.__empty_stockpile())



"""
The module with CLI code that handles the `hopla config` command.
"""

import logging
from pathlib import Path

import click

from hopla.hoplalib.configuration import ConfigurationFileParser
from hopla.hoplalib.configuration import HoplaConfigurationFile

log = logging.getLogger()


def config_get(config_name: str):
    """Get the value of the specified full configuration name"""
    if config_name is None:
        click.echo(config.help)
    else:
        config_file_parser = ConfigurationFileParser()
        value = config_file_parser.get_full_config_name(full_config_name=config_name)
        click.echo(f"{value}")


def config_set(config_name: str, value):
    """Set the value of the specified full configuration name"""
    config_file_parser = ConfigurationFileParser()
    value = config_file_parser.set_full_config_name(full_config_name=config_name,
                                                    new_value=value)
    click.echo(f"{config_name}={value}")


def print_config_file_name():
    """
    Prints the path of the configuration file that this `hopla config ...` invocation uses.
    """
    click.echo(HoplaConfigurationFile().file_path)


def print_config_file_content():
    """Print the hopla config file's content."""
    config_file_path: Path = HoplaConfigurationFile().file_path
    with open(config_file_path, mode="r", encoding="utf-8") as conf_file:
        for line in conf_file.readlines():
            click.echo(line.strip())


supported_config_names = click.Choice(["cmd_all.loglevel"])
"""Values supported: debug,info,warning,error"""


@click.command()
@click.argument("config_name", required=False, type=supported_config_names)
@click.argument("value", required=False)
@click.option("--list/--no-flag", "list_flag",
              default=False, show_default=True,
              help="List the contents of the config file.")
@click.option("--config-file-name/--no-config-file-name", "config_file_name_flag",
              default=False, show_default=True,
              help="Get the absolute path of the config file.")
def config(config_name: str,
           value,
           list_flag: bool,
           config_file_name_flag: bool):
    """Get, set, or list config values.

    Note that hopla config emulates git config in structure

    \b
    hopla config CONFIG_NAME        # get value of CONFIG_NAME
    hopla config CONFIG_NAME VALUE  # give CONFIG_NAME the specified VALUE
    hopla config --list             # show all config values

    \b
    Examples
    ---
    # get logging level
    $ hopla config cmd_all.loglevel
    warning

    \b
    # set logging level to info
    $ hopla config cmd_all.loglevel info
    cmd_all.loglevel=info

    \b
    # list the configuration values
    $ hopla config --list

    \b
    # display the config file path
    $ hopla config --config-file-name


    \f
    :param list_flag: whether to display the config file name
    :param config_file_name_flag: whether to display the config file name
    :param value:
    :param config_name:
    :return:
    """
    log.debug(f"hopla config name={config_name} value={value}")
    log.debug(f" options: --list={list_flag} --config_file_name={config_file_name_flag}")

    if list_flag:
        print_config_file_content()
    elif config_file_name_flag:
        print_config_file_name()
    elif value is None:
        config_get(config_name)
    else:
        config_set(config_name, value)
